from pathlib import Path

import pytest

from fpdf import FPDF, TextStyle, TitleStyle, errors
from fpdf.enums import Align
from fpdf.outline import TableOfContents

from test.conftest import LOREM_IPSUM, assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_simple_outline(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()
    pdf.set_y(50)
    pdf.set_font(size=40)
    p(pdf, "Doc Title", align="C")
    pdf.set_font(size=12)
    pdf.insert_toc_placeholder(render_toc)
    insert_test_content(pdf)
    assert_pdf_equal(pdf, HERE / "simple_outline.pdf", tmp_path)


def render_toc(pdf, outline):
    pdf.y += 50
    pdf.set_font("Helvetica", size=16)
    pdf.underline = True
    p(pdf, "Table of contents:")
    pdf.underline = False
    pdf.y += 20
    pdf.set_font("Courier", size=12)
    for section in outline:
        link = pdf.add_link(page=section.page_number)
        p(
            pdf,
            f'{" " * section.level * 2} {section.name} {"." * (60 - section.level*2 - len(section.name))} {section.page_number}',
            align="C",
            link=link,
        )


def test_insert_toc_placeholder_with_invalid_arg_type():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(TypeError):
        pdf.insert_toc_placeholder("render_toc")


def test_insert_toc_placeholder_twice():
    pdf = FPDF()
    pdf.add_page()
    pdf.insert_toc_placeholder(render_toc)
    with pytest.raises(errors.FPDFException):
        pdf.insert_toc_placeholder(render_toc)


def test_incoherent_start_section_hierarchy():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(ValueError):
        pdf.start_section("Title", level=-1)
    pdf.start_section("Title", level=0)
    with pytest.raises(ValueError):
        pdf.start_section("Subtitle", level=2)


def test_start_section_horizontal_alignment(tmp_path):  # issue-1282
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=20)

    # left align
    level0 = TextStyle(font_size_pt=20, l_margin=Align.L)
    pdf.set_section_title_styles(level0)
    pdf.start_section("left aligned section")

    # center align
    level0 = TextStyle(font_size_pt=20, l_margin=Align.C)
    pdf.set_section_title_styles(level0)
    pdf.start_section("center aligned section")

    # right align
    level0 = TextStyle(font_size_pt=20, l_margin=Align.R)
    pdf.set_section_title_styles(level0)
    pdf.start_section("right aligned section")

    assert_pdf_equal(
        pdf, HERE / "test_start_section_horizontal_alignment.pdf", tmp_path
    )


def test_set_section_title_styles_with_invalid_arg_type():
    pdf = FPDF()
    with pytest.raises(TypeError):
        pdf.set_section_title_styles("Times")


def test_2_pages_outline(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica")
    pdf.set_section_title_styles(
        # Level 0 titles:
        TextStyle(
            font_family="Times",
            font_style="B",
            font_size_pt=24,
            color=128,
            underline=True,
            t_margin=10,
            l_margin=10,
            b_margin=0,
        ),
    )

    pdf.add_page()
    pdf.set_y(50)
    pdf.set_font(size=40)
    p(pdf, "Doc Title", align="C")
    pdf.set_font(size=12)
    pdf.insert_toc_placeholder(render_toc, pages=2)
    for i in range(40):
        pdf.start_section(f"Title {i}")
        p(
            pdf,
            (
                "Lorem ipsum dolor sit amet, consectetur adipiscing elit,"
                " sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
            ),
        )
    assert_pdf_equal(pdf, HERE / "2_pages_outline.pdf", tmp_path)


def test_2_pages_outline_with_deprecated_TitleStyle(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica")
    with pytest.warns(DeprecationWarning):
        pdf.set_section_title_styles(
            # Level 0 titles:
            TitleStyle(
                font_family="Times",
                font_style="B",
                font_size_pt=24,
                color=128,
                underline=True,
                t_margin=10,
                l_margin=10,
                b_margin=0,
            ),
        )

    pdf.add_page()
    pdf.set_y(50)
    pdf.set_font(size=40)
    p(pdf, "Doc Title", align="C")
    pdf.set_font(size=12)
    pdf.insert_toc_placeholder(render_toc, pages=2)
    for i in range(40):
        pdf.start_section(f"Title {i}")
        p(
            pdf,
            (
                "Lorem ipsum dolor sit amet, consectetur adipiscing elit,"
                " sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
            ),
        )
    assert_pdf_equal(pdf, HERE / "2_pages_outline.pdf", tmp_path)


def test_toc_with_nb_and_footer(tmp_path):  # issue-548
    class TestPDF(FPDF):
        def render_toc(self, outline):
            self.x = self.l_margin
            self.set_font(size=12)
            for section in outline:
                self.ln()
                self.cell(text=section.name)

        def footer(self):
            self.set_y(-15)
            self.set_font("helvetica", style="I", size=8)
            self.cell(0, 10, f"Page {self.page_no()}/{{nb}}", align="C")

    pdf = TestPDF()
    pdf.set_font(family="helvetica", size=12)
    pdf.add_page()
    pdf.insert_toc_placeholder(TestPDF.render_toc, pages=2)
    for i in range(1, 80):
        pdf.set_font(style="B")
        pdf.start_section(f"Section {i}")
        pdf.cell(text=f"Section {i}")
        pdf.ln()

    assert_pdf_equal(pdf, HERE / "toc_with_nb_and_footer.pdf", tmp_path)


def test_toc_with_russian_heading(tmp_path):  # issue-320
    pdf = FPDF()
    pdf.add_font(fname="test/fonts/Roboto-Regular.ttf")
    pdf.set_font("Roboto-Regular")
    pdf.add_page()
    pdf.start_section("Русский, English, 1 2 3...")
    pdf.write(8, "Русский текст в параграфе.")
    assert_pdf_equal(pdf, HERE / "toc_with_russian_heading.pdf", tmp_path)


def test_toc_with_thai_headings(tmp_path):  # issue-458
    pdf = FPDF()
    for txt in [
        "ลักษณะเฉพาะของคุณ",
        "ระดับฮอร์โมนเพศชาย",
        "ระดับฮอร์โมนเพศหญิง",
        "hello",
    ]:
        pdf.add_page()
        pdf.start_section(txt)
    assert_pdf_equal(pdf, HERE / "toc_with_thai_headings.pdf", tmp_path)


def test_toc_without_font_style(tmp_path):  # issue-676
    pdf = FPDF()
    pdf.set_font("helvetica")
    pdf.set_section_title_styles(
        level0=TextStyle(font_size_pt=28, l_margin=10), level1=TextStyle()
    )
    pdf.add_page()
    pdf.start_section("Title")
    pdf.start_section("Subtitle", level=1)
    assert_pdf_equal(pdf, HERE / "toc_without_font_style.pdf", tmp_path)


def test_toc_with_font_style_override_bold(tmp_path):  # issue-1072
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", style="B")
    pdf.set_section_title_styles(TextStyle("Helvetica", font_size_pt=20))
    pdf.start_section("foo")
    assert_pdf_equal(pdf, HERE / "toc_with_font_style_override_bold1.pdf", tmp_path)

    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", style="B")
    pdf.set_section_title_styles(TextStyle("Helvetica", font_size_pt=20, font_style=""))
    pdf.start_section("foo")
    assert_pdf_equal(pdf, HERE / "toc_with_font_style_override_bold2.pdf", tmp_path)


def test_toc_without_font_style_with_deprecated_TitleStyle(tmp_path):
    pdf = FPDF()
    pdf.set_font("helvetica")
    with pytest.warns(DeprecationWarning):
        pdf.set_section_title_styles(
            level0=TitleStyle(font_size_pt=28, l_margin=10), level1=TitleStyle()
        )
    pdf.add_page()
    pdf.start_section("Title")
    pdf.start_section("Subtitle", level=1)
    assert_pdf_equal(pdf, HERE / "toc_without_font_style.pdf", tmp_path)


def test_toc_with_font_style_override_bold_with_deprecated_TitleStyle(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", style="B")
    with pytest.warns(DeprecationWarning):
        pdf.set_section_title_styles(
            TitleStyle("Helvetica", font_size_pt=20, color=(0, 0, 0))
        )
    pdf.start_section("foo")
    assert_pdf_equal(pdf, HERE / "toc_with_font_style_override_bold1.pdf", tmp_path)

    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", style="B")
    with pytest.warns(DeprecationWarning):
        pdf.set_section_title_styles(
            TitleStyle("Helvetica", font_style="", font_size_pt=20, color=(0, 0, 0))
        )
    pdf.start_section("foo")
    assert_pdf_equal(pdf, HERE / "toc_with_font_style_override_bold2.pdf", tmp_path)


def test_toc_with_table(tmp_path):  # issue-1079
    def render_toc_with_table(pdf: FPDF, outline: list):
        pdf.set_font(size=20)
        with pdf.table([[x.name, str(x.page_number)] for x in outline]):
            pass

    pdf = FPDF()
    pdf.set_font(family="helvetica", size=30)
    pdf.add_page()
    pdf.insert_toc_placeholder(render_toc_function=render_toc_with_table, pages=4)
    for i in range(60):
        pdf.start_section(name=str(i), level=0)
        pdf.cell(text=str(i))
        pdf.ln()
    assert_pdf_equal(pdf, HERE / "toc_with_table.pdf", tmp_path)


def test_toc_with_right_aligned_page_numbers(tmp_path):
    def render_toc_with_right_aligned_page_numbers(pdf, outline):
        pdf.set_font("Helvetica", size=16)
        for section in outline:
            link = pdf.add_link(page=section.page_number)
            pdf.cell(
                text=f'{" " * section.level * 2} {section.name}',
                link=link,
                new_x="LEFT",
            )
            pdf.cell(text=f"{section.page_number}", link=link, w=pdf.epw, align="R")
            pdf.ln()

    pdf = FPDF()
    pdf.set_font("Helvetica", size=12)
    pdf.add_page()
    pdf.insert_toc_placeholder(render_toc_with_right_aligned_page_numbers)
    insert_test_content(pdf)
    assert_pdf_equal(pdf, HERE / "toc_with_right_aligned_page_numbers.pdf", tmp_path)


def p(pdf, text, **kwargs):
    "Inserts a paragraph"
    pdf.multi_cell(
        w=pdf.epw,
        h=pdf.font_size,
        text=text,
        new_x="LMARGIN",
        new_y="NEXT",
        **kwargs,
    )


def insert_test_content(pdf):
    pdf.set_section_title_styles(
        # Level 0 titles:
        TextStyle(
            font_family="Times",
            font_style="B",
            font_size_pt=24,
            color=128,
            underline=True,
            t_margin=10,
            l_margin=10,
            b_margin=0,
        ),
        # Level 1 subtitles:
        TextStyle(
            font_family="Times",
            font_style="B",
            font_size_pt=20,
            color=128,
            underline=True,
            t_margin=10,
            l_margin=20,
            b_margin=5,
        ),
    )

    pdf.start_section("Title 1")
    pdf.start_section("Subtitle 1.1", level=1)
    p(
        pdf,
        (
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit,"
            " sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
        ),
    )
    pdf.add_page()
    pdf.start_section("Subtitle 1.2", level=1)
    p(
        pdf,
        "Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.",
    )
    pdf.add_page()
    pdf.start_section("Title 2")
    pdf.start_section("Subtitle 2.1", level=1)
    p(
        pdf,
        "Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur.",
    )
    pdf.add_page()
    pdf.start_section("Subtitle 2.2", level=1)
    p(
        pdf,
        "Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.",
    )


def test_insert_toc_placeholder_with_last_page_in_landscape(tmp_path):
    pdf = FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()
    pdf.set_y(50)
    pdf.set_font(size=40)
    p(pdf, "Doc Title", align="C")
    pdf.set_font(size=12)
    pdf.insert_toc_placeholder(render_toc)
    insert_test_content(pdf)
    pdf.add_page(orientation="L")
    pdf.start_section("Title 3")
    p(pdf, text="Nullam tempus magna quam, ac dictum neque blandit quis.")
    assert_pdf_equal(
        pdf, HERE / "insert_toc_placeholder_with_last_page_in_landscape.pdf", tmp_path
    )


@pytest.mark.parametrize(
    "test_number",
    tuple(range(3)),
)
def test_toc_extra_pages_with_labels(tmp_path, test_number):
    file_content = [
        {
            "name": "Chapter 1: The Origins of Procrastination",
            "level": 0,
            "paragraphs": 1,
        },
        {
            "name": "1.1 A Brief History of Putting Things Off",
            "level": 1,
            "paragraphs": 3,
        },
        {
            "name": "1.1.1 Ancient Procrastinators: Pharaohs and Philosophers",
            "level": 2,
            "paragraphs": 3,
        },
        {
            "name": "1.1.2 Renaissance Slacking: Michelangelo's 4-Year Ceiling Break",
            "level": 2,
            "paragraphs": 2,
        },
        {
            "name": "1.2 Famous Procrastinators Through Time",
            "level": 1,
            "paragraphs": 2,
        },
        {"name": "1.2.1 The Tale of Last-Minute Da Vinci", "level": 2, "paragraphs": 3},
        {
            "name": "1.2.2 Procrastination in the Age of the Internet",
            "level": 2,
            "paragraphs": 4,
        },
        {
            "name": "Chapter 2: The Science Behind Delaying Everything",
            "level": 0,
            "paragraphs": 1,
        },
        {"name": "2.1 Why We Procrastinate", "level": 1, "paragraphs": 2},
        {
            "name": "2.1.1 The Pleasure Principle vs. Deadline Panic",
            "level": 2,
            "paragraphs": 2,
        },
        {
            "name": "2.1.2 Brain Chemistry: Dopamine's Role in Putting Off Tasks",
            "level": 2,
            "paragraphs": 3,
        },
        {"name": "2.2 The Types of Procrastinators", "level": 1, "paragraphs": 2},
        {"name": "2.2.1 The Creative Avoider", "level": 2, "paragraphs": 4},
        {"name": "2.2.2 The Perfectionist Delayer", "level": 2, "paragraphs": 2},
        {"name": "2.2.3 The Professional Postponer", "level": 2, "paragraphs": 3},
        {
            "name": "Chapter 3: Procrastination Techniques and Mastery",
            "level": 0,
            "paragraphs": 1,
        },
        {
            "name": "3.1 Strategies for Effective Procrastination",
            "level": 1,
            "paragraphs": 2,
        },
        {
            "name": '3.1.1 The 10-Minute "Break" That Lasts an Hour',
            "level": 2,
            "paragraphs": 1,
        },
        {
            "name": "3.1.2 Productive Procrastination: Cleaning When You Should be Working",
            "level": 2,
            "paragraphs": 2,
        },
        {"name": "3.2 Advanced Delaying Techniques", "level": 1, "paragraphs": 1},
        {
            "name": '3.2.1 The "I\'ll Start After Lunch" Maneuver',
            "level": 2,
            "paragraphs": 2,
        },
        {"name": "3.2.2 The Art of Endless List-Making", "level": 2, "paragraphs": 3},
        {"name": "3.2.3 The Social Media Rabbit Hole", "level": 2, "paragraphs": 5},
        {
            "name": "Chapter 4: Consequences of Procrastination (That We'll Worry About Later)",
            "level": 0,
            "paragraphs": 1,
        },
        {
            "name": "4.1 What Happens When You Put Things Off",
            "level": 1,
            "paragraphs": 2,
        },
        {"name": "4.1.1 Stress Levels and Sudden Panics", "level": 2, "paragraphs": 3},
        {
            "name": "4.1.2 Creative Excuses for Deadline Extensions",
            "level": 2,
            "paragraphs": 4,
        },
        {
            "name": "4.2 Real-World Examples of Procrastination Fails",
            "level": 1,
            "paragraphs": 3,
        },
        {"name": "4.2.1 The Last-Minute Report Disaster", "level": 2, "paragraphs": 4},
        {"name": "4.2.2 The All-Nighter Presentation", "level": 2, "paragraphs": 3},
        {
            "name": "Chapter 5: Embracing Procrastination as a Lifestyle",
            "level": 0,
            "paragraphs": 1,
        },
        {"name": "5.1 Procrastination as a Philosophy", "level": 1, "paragraphs": 2},
        {"name": "5.1.1 The Zen of Doing It Tomorrow", "level": 2, "paragraphs": 2},
        {"name": "5.1.2 Practicing Mindful Delay", "level": 2, "paragraphs": 3},
        {
            "name": (
                "5.1.3 An Incredibly Long and Detailed Guide to Procrastination Tools and Techniques "
                "that Every Expert Procrastinator Should Know, Including but Not Limited to Avoidance Strategies, "
                "Deadline Extensions, and Perfecting the Art of Doing Absolutely Nothing While Looking Busy"
            ),
            "level": 2,
            "paragraphs": 5,
        },
        {"name": "5.2 The Future of Procrastination", "level": 1, "paragraphs": 2},
        {"name": "5.2.1 Delaying with AI Assistance", "level": 2, "paragraphs": 3},
        {
            "name": "5.2.2 The Rise of Procrastination Apps and Tools",
            "level": 2,
            "paragraphs": 2,
        },
    ]

    def footer():
        if pdf.page == 1:
            return
        pdf.set_y(pdf.h - 10)
        pdf.cell(text=pdf.get_page_label(), center=True)

    pdf = FPDF()
    pdf.footer = footer

    pdf.add_page()
    pdf.set_font("helvetica", "", 60)
    pdf.cell(w=pdf.epw, text="TITLE", align="C")

    pdf.set_font("helvetica", "", 12)

    pdf.set_section_title_styles(
        level0=TextStyle(
            font_family="helvetica",
            font_style="B",
            font_size_pt=16,
            color="#00316e",
            b_margin=10,
        ),
        level1=TextStyle(
            font_family="helvetica",
            font_style="B",
            font_size_pt=14,
            t_margin=2.5,
            b_margin=5,
        ),
        level2=TextStyle(
            font_family="helvetica",
            font_style="I",
            font_size_pt=14,
            t_margin=2.5,
            b_margin=5,
        ),
    )

    if test_number == 0:
        # Test with a standard table of contents
        pdf.add_page(label_style="r")
        pdf.start_section("Table of Contents 1", level=0)
        toc = TableOfContents()
        pdf.insert_toc_placeholder(toc.render_toc, allow_extra_pages=True)

    if test_number == 1:
        # Test with a ToC that has a different style using a ttf font
        pdf.add_page(label_style="R")
        pdf.start_section("Table of Contents 2", level=0)
        pdf.multi_cell(
            w=pdf.epw,
            text="This is a second table of contents where a custom font is used for the ToC only.",
            new_x="lmargin",
            new_y="next",
        )
        pdf.ln()
        pdf.multi_cell(
            w=pdf.epw,
            text="Also adding extra lines to test the ToC starting renderening from the position the function is invoked.",
            new_x="lmargin",
            new_y="next",
        )
        pdf.ln()
        pdf.ln()
        pdf.add_font(
            family="Quicksand",
            fname=HERE.parent / "fonts" / "Quicksand-Regular.otf",
        )
        toc = TableOfContents()
        toc.text_style = TextStyle(font_family="Quicksand", font_size_pt=14)
        pdf.insert_toc_placeholder(toc.render_toc, allow_extra_pages=True)

    if test_number == 2:
        # Test using the section styles on the ToC
        pdf.add_page(label_style="a")
        pdf.start_section("Table of Contents 3", level=0)
        pdf.multi_cell(
            w=pdf.epw,
            text="This is a third table of contents using the same style as the sections.",
            new_x="lmargin",
            new_y="next",
        )
        pdf.ln()
        pdf.ln()
        toc = TableOfContents(use_section_title_styles=True)
        pdf.insert_toc_placeholder(toc.render_toc, allow_extra_pages=True)

    pdf.set_page_label(label_style="D")
    for index, content in enumerate(file_content):
        if content["level"] == 0 and index > 0:
            pdf.add_page()
        pdf.start_section(content["name"], level=content["level"])
        for _ in range(content["paragraphs"]):
            pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM, new_x="lmargin", new_y="next")
            pdf.ln()

    assert_pdf_equal(pdf, HERE / f"toc_with_extra_page_{test_number}.pdf", tmp_path)


def test_page_label():
    pdf = FPDF()

    # Test with no label style, prefix, or starting number (default behavior)
    pdf.add_page(label_style=None, label_prefix=None, label_start=None)
    assert pdf.get_page_label() == "1"
    pdf.add_page()
    assert pdf.get_page_label() == "2"

    # Test decimal numerals with no prefix and starting from 1
    pdf.add_page(label_style="D", label_prefix=None, label_start=1)
    assert pdf.get_page_label() == "1"
    pdf.add_page()
    assert pdf.get_page_label() == "2"

    # Test decimal numerals with a prefix and starting from 67
    pdf.add_page(label_style="D", label_prefix="A-", label_start=67)
    assert pdf.get_page_label() == "A-67"
    pdf.add_page()
    assert pdf.get_page_label() == "A-68"

    # Test lowercase Roman numerals starting from 1
    pdf.add_page(label_style="r", label_prefix=None, label_start=1)
    assert pdf.get_page_label() == "i"
    pdf.add_page()
    assert pdf.get_page_label() == "ii"

    # Test uppercase Roman numerals starting from 10 with prefix
    pdf.add_page(label_style="R", label_prefix="Preface-", label_start=10)
    assert pdf.get_page_label() == "Preface-X"
    pdf.add_page()
    assert pdf.get_page_label() == "Preface-XI"

    # Test uppercase letters starting from 1
    pdf.add_page(label_style="A", label_prefix=None, label_start=1)
    assert pdf.get_page_label() == "A"
    pdf.add_page()
    assert pdf.get_page_label() == "B"
    pdf.add_page()
    assert pdf.get_page_label() == "C"

    # Test uppercase letters with a prefix and starting beyond Z
    pdf.add_page(label_style="A", label_prefix="Appendix-", label_start=26)
    assert pdf.get_page_label() == "Appendix-Z"
    pdf.add_page()
    assert pdf.get_page_label() == "Appendix-AA"
    pdf.add_page()
    assert pdf.get_page_label() == "Appendix-AB"

    # Test lowercase letters starting from 1
    pdf.add_page(label_style="a", label_prefix=None, label_start=1)
    assert pdf.get_page_label() == "a"
    pdf.add_page()
    assert pdf.get_page_label() == "b"
    pdf.add_page()
    assert pdf.get_page_label() == "c"

    # Test lowercase letters with prefix and starting beyond z
    pdf.add_page(label_style="a", label_prefix="Sec-", label_start=26)
    assert pdf.get_page_label() == "Sec-z"
    pdf.add_page()
    assert pdf.get_page_label() == "Sec-aa"
    pdf.add_page()
    assert pdf.get_page_label() == "Sec-ab"


def test_toc_no_reset_page_indices(tmp_path):  # cf. issue 1343
    class CustomFPDF(FPDF):
        def footer(self):
            self.set_y(-15)
            # The page number inserted below is incorrect there :(
            p(pdf, "Page %s of {nb}" % self.get_page_label(), align="C")

    pdf = CustomFPDF()
    pdf.set_font("Helvetica", size=12)
    pdf.set_section_title_styles(TextStyle())  # Use default style
    pdf.add_page()
    pdf.set_page_label(label_style="R")
    p(pdf, "**Document title**", align="C", markdown=True)
    pdf.add_page()
    pdf.start_section("Table of Content")
    pdf.insert_toc_placeholder(
        TableOfContents().render_toc, allow_extra_pages=True, reset_page_indices=False
    )
    pdf.set_page_label(label_style="D")
    assert pdf.pages_count == 3
    page_label = pdf.pages[pdf.page].get_page_label()
    assert page_label.get_start() == 3

    for i in range(1, 100):
        if i > 1:
            pdf.add_page()
        pdf.start_section(f"Section {i}")

    assert_pdf_equal(pdf, HERE / "toc_no_reset_page_indices.pdf", tmp_path)


def test_footer_leaking_style_on_toc(tmp_path):
    class CustomFPDF(FPDF):
        def footer(self):
            self.set_y(-15)
            self.set_font("Helvetica", size=8)
            self.cell(w=0, text=f"PAGE {self.get_page_label()}", align="C")

    pdf = CustomFPDF()
    pdf.set_font("Helvetica", size=16)
    pdf.set_section_title_styles(TextStyle())  # Use default style
    pdf.add_page()
    pdf.set_page_label(label_style="R")
    pdf.insert_toc_placeholder(TableOfContents().render_toc, allow_extra_pages=True)
    pdf.set_page_label(label_style="D")
    for i in range(1, 100):
        if i > 1:
            pdf.add_page()
        pdf.start_section(f"Section {i}")
    assert_pdf_equal(pdf, HERE / "footer_leaking_style_on_toc.pdf", tmp_path)



import math

from fpdf.util import convert_unit


def test_convert_unit_number_number():
    """Test converting units where both old and new are numbers"""
    converted = convert_unit(1, 72, 72 / 25.4)  # inch to mm
    assert math.isclose(converted, 25.4), "1 inch should equal 25.4 mm at 72dpi"


def test_convert_unit_number_string():
    """Test converting units where old is a number and new is a string"""
    converted = convert_unit(1, 72, "mm")  # inch to mm
    assert math.isclose(converted, 25.4), "1 inch should equal 25.4 mm at 72dpi"


def test_convert_unit_string_number():
    """Test converting units where old is a string and new is a number"""
    converted = convert_unit(1, "in", 72 / 25.4)  # inch to mm
    assert math.isclose(converted, 25.4), "1 inch should equal 25.4 mm at 72dpi"


def test_convert_unit_string_string():
    """Test converting units where both old and new are strings"""
    converted = convert_unit(1, "in", "mm")
    assert math.isclose(converted, 25.4), "1 inch should equal 25.4 mm at 72dpi"


def test_convert_unit_tuple():
    """
    Test converting a simple tuple

    Eg. The output of `get_page_format("letter", 1)`
    """
    inches = convert_unit((612, 792), "pt", "in")
    assert inches == (8.5, 11.0)


def test_convert_unit_list():
    """Test converting a simple list"""
    inches = convert_unit([612, 792], "pt", "in")
    assert inches == (8.5, 11.0)


def test_convert_unit_iterator():
    """Test converting a simple iterator"""
    inches = convert_unit(iter([612, 792]), "pt", "in")
    assert inches == (8.5, 11.0)


def test_convert_unit_list_of_points():
    """Test converting a list of points"""
    inches = convert_unit([(72, 72), (612, 792)], "pt", "in")
    assert inches == ((1, 1), (8.5, 11.0))


def test_convert_unit_crazy():
    """
    Test converting a tuple which holds a list/iterator/tuple which contains points.
    This would be creating points for multiple pdfs at once, where each page of each pdf has its own unique point grid.
    """
    to_convert = (
        [iter([8.5, 11]), [1, 1]],
        ((1, 1), (8.5, 11)),
        iter([(2, 2), (11, 8.5)]),
    )
    converted = convert_unit(to_convert, "in", "mm")
    assert converted == (
        ((215.89999999999998, 279.4), (25.399999999999995, 25.399999999999995)),
        ((25.399999999999995, 25.399999999999995), (215.89999999999998, 279.4)),
        ((50.79999999999999, 50.79999999999999), (279.4, 215.89999999999998)),
    )



from pathlib import Path

import pytest

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def add_cell_page(pdf, cell_text, h=None):
    """Add a page with with some text to the PDF"""
    pdf.set_font("Helvetica")
    pdf.add_page()
    pdf.cell(w=10, h=h or 10, text=str(cell_text))


def test_unit_default(tmp_path):
    """Test creating a PDF with no unit"""
    pdf = FPDF()
    add_cell_page(pdf, "default")
    assert_pdf_equal(pdf, HERE / "unit_default.pdf", tmp_path)


@pytest.mark.parametrize("unit", ["pt", "mm", "cm", "in"])
def test_unit_str(tmp_path, unit):
    """Test all of the unit strings"""
    pdf = FPDF(unit=unit)
    add_cell_page(pdf, unit)
    assert_pdf_equal(pdf, HERE / f"unit_{unit}.pdf", tmp_path)


def test_unit_int(tmp_path):
    """Test creating a PDF with an int unit"""
    pdf = FPDF(unit=144)
    add_cell_page(pdf, "int = 2 in", h=5)
    assert_pdf_equal(pdf, HERE / "unit_int.pdf", tmp_path)


def test_unit_float(tmp_path):
    """Test creating a PDF with a float unit  (same as "in"), and that it is created exactly the same"""
    pdf = FPDF(unit=0.072)
    add_cell_page(pdf, "float = 1 thou")
    assert_pdf_equal(pdf, HERE / "unit_float.pdf", tmp_path)


def test_unit_int_matches_float_and_str(tmp_path):
    """Pdfs created with `unit=<float>` or `unit=<int>` should match those created with the associated string"""
    int_pdf = FPDF(unit=72)
    add_cell_page(int_pdf, "in")
    float_pdf = FPDF(unit=72.0)
    add_cell_page(float_pdf, "in")
    pdf_str = FPDF(unit="in")
    add_cell_page(pdf_str, "in")
    assert_pdf_equal(int_pdf, float_pdf, tmp_path)
    assert_pdf_equal(int_pdf, pdf_str, tmp_path)
    assert_pdf_equal(float_pdf, pdf_str, tmp_path)



from pathlib import Path
from pytest import raises

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_dash_pattern(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=10)

    def draw_stuff(x, y):
        pdf.line(x, y, x + 50, y + 50)
        pdf.polyline(((x, y), (x + 40, y + 20), (x + 10, y + 30), (x + 50, y + 50)))
        pdf.polygon(((x + 5, y + 20), (x + 25, y + 45), (x + 40, y + 10)))
        pdf.rect(x, y, 50, 50)
        pdf.ellipse(x, y, 50, 50)
        pdf.set_xy(x, y + 55)
        pdf.cell(w=50, h=5, text="cell", border=1)

    # solid line
    draw_stuff(20, 20)
    # simple dash
    pdf.set_dash_pattern(3)
    draw_stuff(100, 20)
    # dashdot by overlap
    pdf.set_dash_pattern(4, 6)
    draw_stuff(20, 100)
    pdf.set_dash_pattern(0.5, 9.5, 3.25)
    # coverage: repeating the same pattern should not add it again
    pdf.set_dash_pattern(0.5, 9.5, 3.25)
    draw_stuff(20, 100)
    # reset to solid
    pdf.set_dash_pattern()
    draw_stuff(100, 100)

    assert_pdf_equal(pdf, HERE / "dash_pattern.pdf", tmp_path)


def test_dash_pattern_badinput():
    pdf = fpdf.FPDF()
    pdf.add_page()
    with raises(ValueError):
        pdf.set_dash_pattern(dash=-1)
    with raises(ValueError):
        pdf.set_dash_pattern(gap=-1)
    with raises(ValueError):
        pdf.set_dash_pattern(phase=-1)
    with raises(ValueError):
        pdf.set_dash_pattern(dash="yo")
    with raises(ValueError):
        pdf.set_dash_pattern(gap="hu")
    with raises(ValueError):
        pdf.set_dash_pattern(phase=None)



import io
from pathlib import Path

from PIL import Image

from fpdf import FPDF, drawing
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
imgpath = str(HERE / "../image/image_types/insert_images_insert_jpg.jpg")


def add_pages_with_rgb_tuple_background(pdf, fill_set=None):
    """
    sets the background using a tuple representing an RGB color and adds a page with text and a small rectangle,
    then writes multiline-text to trigger a page break to test if the background color is retained
    """
    pdf.set_page_background((170, 200, 100))
    pdf.add_page()
    pdf.rect(20, 20, 60, 60, style="F")
    pdf.cell(0, 30, "RGB tuple", align="R", border=1)
    pdf.ln()
    if fill_set:
        pdf.cell(0, 30, "fill color set", align="R", fill=True)
    else:
        pdf.cell(0, 30, "default fill color", align="R")
    pdf.ln()
    pdf.multi_cell(0, pdf.h / 2, "PAGE 1\nPAGE 2")


def add_page_with_DeviceRGB_background(pdf):
    """sets the background using an instance of DeviceRGB and adds a page with text and a small rectangle"""
    pdf.set_page_background(drawing.DeviceRGB(0.2, 0.4, 0.1))
    pdf.add_page()
    pdf.rect(20, 20, 60, 60, style="F")
    pdf.cell(0, 10, "DeviceRGB", align="R", border=1)


def _add_pages_with_image_background(pdf):
    """
    sets the background to an image specified with a path and adds a page with text and a small rectangle,
    then writes multi-line text to trigger a page break to test if the background image is retained
    """
    pdf.set_page_background(imgpath)
    pdf.add_page()
    pdf.rect(20, 20, 60, 60, style="F")
    pdf.cell(0, 10, "Image file path", align="R", border=1)
    pdf.ln()
    pdf.multi_cell(0, pdf.h / 2, "PAGE 1\nPAGE 2")


def add_page_with_image_url_background(pdf):
    """sets the background to an image specified with a link and adds a page with text and a small rectangle"""
    pdf.set_page_background(
        "https://raw.githubusercontent.com/py-pdf/fpdf2/master/test/image/image_types/insert_images_insert_jpg.jpg"
    )
    pdf.add_page()
    pdf.rect(20, 20, 60, 60, style="F")
    pdf.cell(0, 10, "Image file link", align="R", border=1)


def add_page_with_Pillow_image_background(pdf):
    """sets the background using an instance of PIL.Image.Image and adds a page with text and a small rectangle"""
    img = Image.open(imgpath)
    pdf.set_page_background(img)
    pdf.add_page()
    pdf.rect(20, 20, 60, 60, style="F")
    pdf.cell(0, 10, "PIL.Image.Image", align="R", border=1)


def add_page_with_BytesIO_background(pdf):
    """sets the background using an instance of io.BytesIO of an image and adds a page with text and a small rectangle"""
    with open(imgpath, "rb") as f:
        buffer = io.BytesIO(f.read())
    pdf.set_page_background(buffer)
    pdf.add_page()
    pdf.rect(20, 20, 60, 60, style="F")
    pdf.cell(0, 10, "io.BytesIO", align="R", border=1)


def add_page_without_background(pdf):
    """sets the background to None (=removes the background) and adds a page with text and a small rectangle"""
    pdf.set_page_background(None)
    pdf.add_page()
    pdf.rect(20, 20, 60, 60, style="F")
    pdf.cell(0, 10, "No background", align="R", border=1)


def test_page_background(tmp_path):
    """
    Test creating a PDF with multiple pages using all pfsible inputs to set a page background,
    drawing a rectangle and writing text on every page, testing if any other color is overwritten,
    writing a multi-line text to test if the background is retained on an automatic page break,
    then setting a fill color, testing if the background color gets overridden and vice versa
    by printing another two pages with a background color and image
    """
    pdf = FPDF()
    pdf.set_font("Helvetica", style="B", size=30)
    add_pages_with_rgb_tuple_background(pdf)
    _add_pages_with_image_background(pdf)
    add_page_with_DeviceRGB_background(pdf)
    add_page_with_image_url_background(pdf)
    add_page_with_Pillow_image_background(pdf)
    add_page_with_BytesIO_background(pdf)
    add_page_without_background(pdf)

    pdf.set_fill_color(255, 200, 210)
    add_pages_with_rgb_tuple_background(pdf, fill_set=True)
    _add_pages_with_image_background(pdf)
    add_page_without_background(pdf)

    assert_pdf_equal(pdf, HERE / "page_background.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF
from test.conftest import LOREM_IPSUM, assert_pdf_equal, EMOJI_TEST_TEXT

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"


def test_gilbert_color(tmp_path):
    # Gilbert Color - Creative Commons license
    # https://github.com/Fontself/TypeWithPride

    pdf = FPDF()
    pdf.add_font("Gilbert", "", HERE / "Gilbert-Color Bold SVG.otf")

    pdf.add_page()
    pdf.set_font("Gilbert", size=16)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.lower(), align="L")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.lower(), align="R")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.lower(), align="J")

    assert_pdf_equal(pdf, HERE / "svg_gilbert.pdf", tmp_path)


def test_svg_bungee(tmp_path):
    # Bungee Color - OFL license
    # https://github.com/djrrb/Bungee

    pdf = FPDF()
    pdf.add_font("Bungee", "", HERE / "BungeeColor-Regular-SVG.ttf")
    pdf.add_page()
    pdf.set_font("Bungee", size=16)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="L")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="R")
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM.upper(), align="J")

    assert_pdf_equal(pdf, HERE / "svg_bungee.pdf", tmp_path)


def test_twitter_emoji_shaping(tmp_path):
    # Twitter Emoji font - MIT license
    # https://github.com/twitter/twemoji
    pdf = FPDF()
    pdf.add_font("TwitterEmoji", "", FONTS_DIR / "TwitterEmoji.ttf")
    pdf.add_page()
    combined_emojis = "🇫🇷 🇺🇸 🇨🇦 🧑 🧑🏽 🧑🏿"
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis without text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("TwitterEmoji", "", 24)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    pdf.ln()
    pdf.set_font("helvetica", "", 24)
    pdf.cell(text="Emojis with text shaping:", new_x="lmargin", new_y="next")
    pdf.set_font("TwitterEmoji", "", 24)
    pdf.set_text_shaping(True)
    pdf.multi_cell(w=pdf.epw, text=combined_emojis, new_x="lmargin", new_y="next")
    assert_pdf_equal(pdf, HERE / "svg_twitter_emoji_shaping.pdf", tmp_path)


def test_twitter_emoji_text(tmp_path):
    text = EMOJI_TEST_TEXT
    pdf = FPDF()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("TwitterEmoji", "", FONTS_DIR / "TwitterEmoji.ttf")
    pdf.set_font("Roboto", "", 24)
    pdf.set_fallback_fonts(["TwitterEmoji"])
    pdf.add_page()
    pdf.multi_cell(w=pdf.epw, text=text)
    assert_pdf_equal(pdf, HERE / "svg_twitter_emoji_text.pdf", tmp_path)



import io
import logging
import sys
from pathlib import Path

from PIL import Image, ImageDraw, ImageFont
import pytest

import fpdf
from fpdf.image_parsing import transcode_monochrome
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_insert_jpg(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.add_page()
    pdf.image(HERE / "insert_images_insert_jpg.jpg", x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_jpg.pdf", tmp_path)


@pytest.mark.skipif(
    sys.platform in ("cygwin", "win32"),
    reason="Required system libraries to generate JPEG2000 images are a PITA to install under Windows",
)
@pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason="JPEG2000 changes were done on Pillow 11 which are not compatible with versions of Python < 3.11",
)
def test_insert_jpg_jpxdecode(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.set_image_filter("JPXDecode")
    pdf.add_page()
    pdf.image(HERE / "insert_images_insert_jpg.jpg", x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_jpg_jpxdecode.pdf", tmp_path)


def test_insert_jpg_flatedecode(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.set_image_filter("FlateDecode")
    pdf.add_page()
    pdf.image(HERE / "insert_images_insert_jpg.jpg", x=15, y=15, h=140)
    if sys.platform in ("cygwin", "win32"):
        # Pillow uses libjpeg-turbo on Windows and libjpeg elsewhere,
        # leading to a slightly different image being parsed and included in the PDF:
        assert_pdf_equal(
            pdf, HERE / "image_types_insert_jpg_flatedecode_windows.pdf", tmp_path
        )
    else:
        assert_pdf_equal(pdf, HERE / "image_types_insert_jpg_flatedecode.pdf", tmp_path)


@pytest.mark.skipif(
    sys.version_info.minor != 13,
    reason="LZWDecode is currently VERY slow, so we want to only execute it ONCE among all Python versions in the test pipeline",
)
def test_insert_jpg_lzwdecode(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.set_image_filter("LZWDecode")
    pdf.add_page()
    pdf.image(HERE / "insert_images_insert_jpg.jpg", x=15, y=15, h=140)
    if sys.platform in ("cygwin", "win32"):
        assert_pdf_equal(
            pdf, HERE / "image_types_insert_jpg_lzwdecode_windows.pdf", tmp_path
        )
    else:
        assert_pdf_equal(pdf, HERE / "image_types_insert_jpg_lzwdecode.pdf", tmp_path)


def test_insert_jpg_cmyk(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.add_page()
    pdf.image(HERE / "insert_images_insert_jpg_cmyk.jpg", x=15, y=15)
    assert_pdf_equal(pdf, HERE / "images_types_insert_jpg_cmyk.pdf", tmp_path)


def test_insert_png(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(HERE / "insert_images_insert_png.png", x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_png.pdf", tmp_path)


def test_insert_png_monochromatic(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(HERE / "../png_test_suite/basi0g01.png", x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_png_monochromatic.pdf", tmp_path)


def test_transcode_monochrome_and_libtiff_support_custom_tags():
    # Fails under WSL on my computer (Lucas), with this error:
    #   AttributeError: module 'PIL._imaging' has no attribute 'libtiff_support_custom_tags'
    # (as a consequence, the test above & everal others also fail, because PDFs don't match)
    # cf. https://github.com/python-pillow/Pillow/issues/7019
    with Image.open("test/image/png_test_suite/basi0g01.png") as img:
        transcode_monochrome(img)


def test_insert_png_alpha(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.add_page()
    pdf.set_font("Helvetica", size=30)
    pdf.cell(w=pdf.epw, h=30, text="BEHIND")
    pdf.image(
        HERE / "../png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png", x=25, y=0, h=40
    )
    assert_pdf_equal(pdf, HERE / "image_types_insert_png_alpha.pdf", tmp_path)


def test_insert_png_disallow_transparency(tmp_path):
    pdf = fpdf.FPDF()
    pdf.allow_images_transparency = False
    pdf.add_page()
    pdf.set_font("Helvetica", size=30)
    pdf.cell(w=pdf.epw, h=30, text="BEHIND")
    pdf.image(
        HERE / "../png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png", x=25, y=0, h=40
    )
    assert_pdf_equal(
        pdf, HERE / "image_types_insert_png_disallow_transparency.pdf", tmp_path
    )


def test_insert_png_alpha_dctdecode(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.set_image_filter("DCTDecode")
    pdf.add_page()
    pdf.image(
        HERE / "../png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png", x=15, y=15, h=140
    )
    if sys.platform in ("cygwin", "win32"):
        # Pillow uses libjpeg-turbo on Windows and libjpeg elsewhere,
        # leading to a slightly different image being parsed and included in the PDF:
        assert_pdf_equal(
            pdf, HERE / "image_types_insert_png_alpha_dctdecode_windows.pdf", tmp_path
        )
    else:
        assert_pdf_equal(
            pdf, HERE / "image_types_insert_png_alpha_dctdecode.pdf", tmp_path
        )


def test_insert_bmp(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.add_page()
    pdf.image(HERE / "circle.bmp", x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_bmp.pdf", tmp_path)


def test_insert_jpg_icc(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page(format=(448, 498))
    pdf.set_margin(0)
    pdf.image(HERE / "insert_images_insert_jpg_icc.jpg", x=0, y=0, h=498)
    # we add the same image a second time to make sure the ICC profile is only included
    # only once in that case
    pdf.add_page(format=(448, 498))
    pdf.image(HERE / "insert_images_insert_jpg_icc.jpg", x=0, y=0, h=498)
    # we add another image with the same ICC profile to make sure it's also included
    # only once in that case
    pdf.add_page(format=(314, 500))
    pdf.image(HERE / "insert_images_insert_jpg_icc_2.jpg", x=0, y=0, h=500)
    assert_pdf_equal(pdf, HERE / "image_types_insert_jpg_icc.pdf", tmp_path)


def test_insert_jpg_invalid_icc(caplog, tmp_path):
    with caplog.at_level(logging.INFO):
        pdf = fpdf.FPDF()
        pdf.add_page(format=(448, 498))
        pdf.set_margin(0)
        pdf.image(HERE / "insert_images_insert_jpg_icc_invalid.jpg", x=0, y=0, h=498)
    assert "Invalid ICC Profile in file" in caplog.text
    assert_pdf_equal(pdf, HERE / "image_types_insert_jpg_icc_invalid.pdf", tmp_path)


def test_insert_gif(tmp_path):
    pdf = fpdf.FPDF()
    pdf.compress = False
    pdf.add_page()
    pdf.image(HERE / "circle.gif", x=15, y=15)
    assert_pdf_equal(pdf, HERE / "image_types_insert_gif.pdf", tmp_path)


def test_insert_g4_tiff(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(HERE / "test.tiff", x=15, y=15)
    assert_pdf_equal(pdf, HERE / "image_types_insert_tiff.pdf", tmp_path)


def test_insert_tiff_cmyk(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.image(HERE / "insert_images_insert_tiff_cmyk.tiff", x=15, y=15)
    assert_pdf_equal(pdf, HERE / "image_types_insert_tiff_cmyk.pdf", tmp_path)


def test_insert_pillow(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    img = Image.open(HERE / "insert_images_insert_png.png")
    pdf.image(img, x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_png.pdf", tmp_path)


def test_insert_pillow_issue_139(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    font = ImageFont.truetype(f"{HERE}/../../fonts/DejaVuSans.ttf", 40)
    for y in range(5):
        for x in range(4):
            img = Image.new(mode="RGB", size=(100, 100), color=(60, 255, 10))
            ImageDraw.Draw(img).text((20, 20), f"{y}{x}", fill="black", font=font)
            pdf.image(img, x=x * 50 + 5, y=y * 50 + 5, w=45)
    assert_pdf_equal(pdf, HERE / "insert_pillow_issue_139.pdf", tmp_path)


def test_insert_bytesio(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    img = Image.open(HERE / "insert_images_insert_png.png")
    img_bytes = io.BytesIO()
    img.save(img_bytes, "PNG")
    pdf.image(img_bytes, x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_png.pdf", tmp_path)
    assert not img_bytes.closed  # cf. issue #881


def test_insert_bytes(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    img = Image.open(HERE / "insert_images_insert_png.png")
    img_bytes = io.BytesIO()
    img.save(img_bytes, "PNG")
    pdf.image(img_bytes.getvalue(), x=15, y=15, h=140)
    assert_pdf_equal(pdf, HERE / "image_types_insert_png.pdf", tmp_path)



# pylint: disable=no-self-use, protected-access
from pathlib import Path

import fpdf
from ..conftest import assert_pdf_equal

from defusedxml.ElementTree import fromstring as parse_xml_str
import pytest


from . import parameters

HERE = Path(__file__).resolve().parent
GENERATED_PDF_DIR = HERE / "generated_pdf"


def assert_style_match(lhs, rhs):
    mismatches = []
    for attr in lhs.MERGE_PROPERTIES:
        left = getattr(lhs, attr)
        right = getattr(rhs, attr)
        # can't short circuit because we want to report all of the issues
        if left != right:
            mismatches.append(f"left.{attr} ({left}) != right.{attr} ({right})")

    if mismatches:
        raise AssertionError("Styles do not match:\n    " + "\n    ".join(mismatches))


class TestUnits:
    @pytest.mark.parametrize(
        "name", [pytest.param(name, id=name) for name in fpdf.svg.relative_length_units]
    )
    def test_resolve_relative_length_units(self, name):
        with pytest.raises(ValueError):
            fpdf.svg.resolve_length(f" 1{name}")

    @pytest.mark.parametrize(
        "name", [pytest.param(name, id=name) for name in fpdf.svg.absolute_length_units]
    )
    def test_resolve_good_length_units(self, name):
        computed = fpdf.svg.resolve_length(f"  1 {name} ")
        assert isinstance(computed, float)

    def test_resolve_bad_length_units(self):
        with pytest.raises(ValueError):
            fpdf.svg.resolve_length("1 fake")

    def test_resolve_implicit_length_units(self):
        value = 1.5
        assert fpdf.svg.resolve_length(f"{value}") == value

    def test_resolve_angle_bad_units(self):
        with pytest.raises(ValueError):
            fpdf.svg.resolve_angle("1 fake")

    @pytest.mark.parametrize(
        "name", [pytest.param(name, id=name) for name in fpdf.svg.angle_units]
    )
    def test_resolve_good_angle_units(self, name):
        computed = fpdf.svg.resolve_angle(f"  1 {name} ")
        assert isinstance(computed, float)


def test_xmlns_lookup_failure():
    result = fpdf.svg.xmlns("this is not a real xml namespace", "rect")
    assert result == "rect"


def test_optional_converter():
    canary = object()
    conv = fpdf.svg.optional(object(), converter=lambda val: canary)
    assert conv is canary


class TestSVGPathParsing:
    @pytest.mark.parametrize("path, result", parameters.svg_path_directives)
    def test_parsing_all_directives(self, path, result):
        pdf_path = fpdf.drawing.PaintedPath()

        fpdf.svg.svg_path_converter(pdf_path, path)

        assert result == pdf_path._root_graphics_context.path_items

    @pytest.mark.parametrize("path, result", parameters.svg_path_implicit_directives)
    def test_parsing_all_implicit_directives(self, path, result):
        pdf_path = fpdf.drawing.PaintedPath()

        fpdf.svg.svg_path_converter(pdf_path, path)

        assert result == pdf_path._root_graphics_context.path_items

    @pytest.mark.parametrize("path, result", parameters.svg_path_edge_cases)
    def test_parsing_edge_cases(self, path, result):
        pdf_path = fpdf.drawing.PaintedPath()

        fpdf.svg.svg_path_converter(pdf_path, path)

        assert result == pdf_path._root_graphics_context.path_items

    def test_bad_path_start(self):
        pdf_path = fpdf.drawing.PaintedPath()
        with pytest.raises(ValueError):
            fpdf.svg.svg_path_converter(pdf_path, "L 1 2")
        with pytest.raises(ValueError):
            fpdf.svg.svg_path_converter(pdf_path, "C 1 2 3 4 5 6")
        with pytest.raises(ValueError):
            fpdf.svg.svg_path_converter(pdf_path, "Q 1 2 3 4")
        with pytest.raises(ValueError):
            fpdf.svg.svg_path_converter(pdf_path, "A 1 2 0 1 0 4 5")

    @pytest.mark.parametrize("path, expected", parameters.svg_path_render_tests)
    def test_rendering_smooth_curves(self, path, expected):
        pdf_path = fpdf.drawing.PaintedPath()

        fpdf.svg.svg_path_converter(pdf_path, path)

        resource_catalog = fpdf.output.ResourceCatalog()
        style = fpdf.drawing.GraphicsStyle()
        first_point = fpdf.drawing.Point(0, 0)
        start = fpdf.drawing.Move(first_point)

        result = pdf_path.render(resource_catalog, style, start, first_point)[0]

        assert result == expected


@pytest.mark.parametrize("shape, output, guard", parameters.test_svg_shape_tags)
def test_svg_shape_conversion(shape, output, guard):
    xml = parse_xml_str(shape)
    converter = getattr(fpdf.svg.ShapeBuilder, xml.tag)

    with guard:
        path = converter(xml)
        assert output == path._root_graphics_context.path_items


class TestSVGAttributeConversion:
    @pytest.mark.parametrize(
        "transform, expected, guard", parameters.test_svg_transforms
    )
    def test_svg_transform_conversion(self, transform, expected, guard):
        with guard:
            result = fpdf.svg.convert_transforms(transform)
            assert result == pytest.approx(expected)

    @pytest.mark.parametrize("svg_file", parameters.test_svg_transform_documents)
    def test_svg_transform_conversion_visual(self, tmp_path, svg_file):
        # this only tests the SVG 1.1 transforms for now because the 2.0 transforms do not seem to be widely supported?
        svg = fpdf.svg.SVGObject.from_file(svg_file)

        pdf = fpdf.FPDF(unit="pt", format=(svg.width, svg.height))
        pdf.set_margin(0)
        pdf.add_page()

        svg.draw_to_page(pdf)

        assert_pdf_equal(
            pdf, GENERATED_PDF_DIR / "transforms" / f"{svg_file.stem}.pdf", tmp_path
        )

    @pytest.mark.parametrize(
        "element, expected, guard", parameters.test_svg_attribute_conversion
    )
    def test_attribute_conversion(self, element, expected, guard):
        xml = parse_xml_str(element)

        stylable = fpdf.drawing.PaintedPath()
        with guard:
            fpdf.svg.apply_styles(stylable, xml)
            assert_style_match(stylable.style, expected)


class TestSVGObject:
    def test_bad_root_tag(self):
        notsvg = """<sometag></sometag>"""

        with pytest.raises(ValueError):
            fpdf.svg.SVGObject(notsvg)

    @pytest.mark.parametrize(
        "svg_data, expected_dim, expected_tf, guard", parameters.svg_shape_info_tests
    )
    def test_document_shape_info(self, svg_data, expected_dim, expected_tf, guard):
        pdf = fpdf.FPDF(unit="pt", format=(10, 10))
        pdf.set_margin(0)
        pdf.add_page()

        with guard:
            svg = fpdf.svg.SVGObject(svg_data)
            width, height, base_group = svg.transform_to_page_viewport(pdf)

            assert (width, height) == pytest.approx(expected_dim)
            assert base_group.transform == pytest.approx(expected_tf)

    @pytest.mark.parametrize("svg_file", parameters.test_svg_sources)
    def test_svg_conversion(self, tmp_path, svg_file):
        svg = fpdf.svg.SVGObject.from_file(svg_file)

        pdf = fpdf.FPDF(unit="pt", format=(svg.width, svg.height))
        pdf.set_margin(0)
        pdf.add_page()

        svg.draw_to_page(pdf)

        assert_pdf_equal(pdf, GENERATED_PDF_DIR / f"{svg_file.stem}.pdf", tmp_path)

    @pytest.mark.parametrize("svg_file", parameters.test_svg_sources[0:1])
    def test_draw_to_page_offset(self, tmp_path, svg_file):
        svg = fpdf.svg.SVGObject.from_file(svg_file)

        pdf = fpdf.FPDF(unit="pt", format=(svg.width, svg.height))
        pdf.set_margin(0)
        pdf.add_page()

        svg.draw_to_page(pdf, x=5, y=5)

        assert_pdf_equal(
            pdf, GENERATED_PDF_DIR / f"{svg_file.stem}-offset.pdf", tmp_path
        )

    def test_path_def(self):
        svg_data = (
            '<svg xmlns="http://www.w3.org/2000/svg" '
            'xmlns:xlink="http://www.w3.org/1999/xlink">'
            '<defs><path id="path" d="M 0 0 L 1 2 Z"/></defs></svg>'
        )
        fpdf.svg.SVGObject(svg_data)

    def test_bad_xref(self):
        svg_data = (
            '<svg xmlns="http://www.w3.org/2000/svg" '
            'xmlns:xlink="http://www.w3.org/1999/xlink">'
            '<use transform="rotate(45)"/></svg>'
        )
        with pytest.raises(ValueError):
            fpdf.svg.SVGObject(svg_data)

    def test_missing_xref(self):
        svg_data = (
            '<svg xmlns="http://www.w3.org/2000/svg" '
            'xmlns:xlink="http://www.w3.org/1999/xlink">'
            '<use xlink:href="#missing"/></svg>'
        )
        with pytest.raises(ValueError):
            fpdf.svg.SVGObject(svg_data)

    def test_svg_conversion_no_transparency(self, tmp_path):
        svg = fpdf.svg.SVGObject.from_file(parameters.svgfile("SVG_logo.svg"))

        pdf = fpdf.FPDF(unit="pt", format=(svg.width, svg.height))
        pdf.set_margin(0)
        pdf.allow_images_transparency = False
        pdf.add_page()

        svg.draw_to_page(pdf)

        assert_pdf_equal(
            pdf, GENERATED_PDF_DIR / "SVG_logo_notransparency.pdf", tmp_path
        )

    def test_svg_conversion_priority_styles(self, tmp_path):
        svg_file = parameters.svgfile("simple_rect.svg")

        svg = fpdf.svg.SVGObject.from_file(svg_file)

        pdf = fpdf.FPDF(unit="pt", format=(svg.width, svg.height))
        pdf.set_margin(0)
        pdf.allow_images_transparency = False
        pdf.add_page()

        svg.draw_to_page(pdf)

        assert_pdf_equal(pdf, GENERATED_PDF_DIR / f"{svg_file.stem}.pdf", tmp_path)

    @pytest.mark.parametrize(
        "fill_color, stroke_color, file_suffix", parameters.svg_current_color
    )
    def test_svg_current_color(self, tmp_path, fill_color, stroke_color, file_suffix):
        svg_file = parameters.svgfile("simple_rect_current_color.svg")

        svg = fpdf.svg.SVGObject.from_file(svg_file)

        pdf = fpdf.FPDF(unit="pt", format=(svg.width, svg.height))
        pdf.set_margin(0)
        pdf.allow_images_transparency = False
        pdf.add_page()

        if fill_color is not None:
            pdf.set_fill_color(fill_color[0], fill_color[1], fill_color[2])
        if stroke_color is not None:
            pdf.set_draw_color(stroke_color[0], stroke_color[1], stroke_color[2])
        svg.draw_to_page(pdf)

        assert_pdf_equal(
            pdf, GENERATED_PDF_DIR / f"{svg_file.stem}{file_suffix}.pdf", tmp_path
        )

    def test_svg_render_content_in_a_tag(self, tmp_path):
        svg_file = parameters.svgfile("simple_rect_in_a_tag.svg")

        svg = fpdf.svg.SVGObject.from_file(svg_file)

        pdf = fpdf.FPDF(unit="pt", format=(svg.width, svg.height))
        pdf.set_margin(0)
        pdf.allow_images_transparency = False
        pdf.add_page()

        svg.draw_to_page(pdf)

        assert_pdf_equal(pdf, GENERATED_PDF_DIR / f"{svg_file.stem}.pdf", tmp_path)

    def test_svg_rendering_image_over_page_break(self, tmp_path):
        pdf = fpdf.FPDF()
        pdf.add_page()
        pdf.y += 150
        pdf.image(parameters.svgfile("embedded-raster-images.svg"))
        # In the resulting document, a page break occurs before the image being rendered:
        assert_pdf_equal(
            pdf, GENERATED_PDF_DIR / "svg_rendering_image_over_page_break.pdf", tmp_path
        )

    def test_svg_text_ttf_font(self, tmp_path):
        pdf = fpdf.FPDF()
        pdf.add_page()
        pdf.add_font(
            family="serif", style="", fname=HERE.parent / "fonts" / "DejaVuSans.ttf"
        )
        pdf.image(
            name=HERE / "svg_sources" / "ocanada.svg",
            x=pdf.l_margin,
            y=pdf.t_margin,
            w=pdf.epw,
            h=pdf.eph,
            keep_aspect_ratio=True,
        )
        assert_pdf_equal(
            pdf,
            GENERATED_PDF_DIR / "ocanada.pdf",
            tmp_path,
        )



from pathlib import Path
from test.conftest import assert_pdf_equal, EPOCH, LOREM_IPSUM

from fpdf import FPDF
from fpdf.recorder import FPDFRecorder

HERE = Path(__file__).resolve().parent


def init_pdf():
    pdf = FPDF()
    pdf.set_creation_date(EPOCH)
    pdf.set_font("helvetica", size=24)
    pdf.add_page()
    pdf.cell(w=pdf.epw, h=10, text="Hello fpdf2!", align="C")
    return pdf


def test_recorder_rewind_ok(tmp_path):
    pdf = init_pdf()
    recorder = FPDFRecorder(pdf)
    expected = recorder.output()  # close the document as a side-effect
    recorder.rewind()  # in order to un-close the document
    recorder.add_page()
    recorder.cell(w=recorder.epw, h=10, text="Hello again!", align="C")
    recorder.rewind()
    assert_pdf_equal(recorder, expected, tmp_path)


def test_recorder_rewind_twice_ok(tmp_path):
    pdf = init_pdf()
    recorder = FPDFRecorder(pdf)
    expected = recorder.output()  # close the document as a side-effect
    recorder.rewind()  # in order to un-close the document
    recorder.add_page()
    recorder.cell(w=recorder.epw, h=10, text="Hello again!", align="C")
    recorder.rewind()
    assert_pdf_equal(recorder, expected, tmp_path)


def test_recorder_replay_ok(tmp_path):
    recorder = FPDFRecorder(init_pdf())
    recorder.add_page()
    recorder.cell(w=recorder.epw, h=10, text="Hello again!", align="C")
    expected = recorder.output()
    recorder.rewind()
    recorder.replay()
    assert_pdf_equal(recorder, expected, tmp_path)


def test_recorder_override_accept_page_break_ok():
    recorder = FPDFRecorder(init_pdf(), accept_page_break=False)
    assert recorder.accept_page_break is False


def test_recorder_preserve_pages_count():
    pdf = init_pdf()
    pdf.set_y(250)
    assert pdf.pages_count == 1
    with pdf.offset_rendering() as recorder:
        pdf.multi_cell(text=LOREM_IPSUM, w=pdf.epw)
        assert pdf.pages_count == 2
    assert recorder.page_break_triggered
    assert pdf.pages_count == 1


def test_recorder_with_ttf_font(tmp_path):
    pdf = init_pdf()
    pdf.add_font(fname=str(HERE / "fonts" / "Roboto-Regular.ttf"))
    pdf.set_font("Roboto-Regular", size=64)
    pdf.add_page()
    pdf.cell(text="Hello!", align="C")
    recorder = FPDFRecorder(pdf)
    expected = recorder.output()  # close the document as a side-effect
    recorder.rewind()  # in order to un-close the document
    recorder.add_page()
    recorder.cell(w=recorder.epw, h=10, text="Hello again!", align="C")
    recorder.rewind()
    assert_pdf_equal(recorder, expected, tmp_path)



from os import devnull
import sys
from pathlib import Path

import pytest

from fpdf import FPDF
from fontTools.ttLib import TTCollection, woff2
from test.conftest import LOREM_IPSUM, assert_pdf_equal, assert_same_file

HERE = Path(__file__).resolve().parent


def test_add_font_non_existing_file():
    pdf = FPDF()
    with pytest.raises(FileNotFoundError) as error:
        pdf.add_font(fname="non-existing-file.ttf")
    assert str(error.value) == "TTF Font file not found: non-existing-file.ttf"


def test_add_font_pkl():
    pdf = FPDF()
    with pytest.raises(ValueError) as error:
        pdf.add_font(fname="non-existing-file.pkl")
    assert str(error.value) == (
        "Unsupported font file extension: .pkl. add_font() used to accept .pkl file as input, "
        "but for security reasons this feature is deprecated since v2.5.1 and has been removed in v2.5.3."
    )


def test_deprecation_warning_for_FPDF_CACHE_DIR_and_FPDF_CACHE_MODE():
    # pylint: disable=import-outside-toplevel,pointless-statement,reimported
    from fpdf import fpdf

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_DIR
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_DIR = "/tmp"
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_MODE
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_MODE = 1
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    fpdf.SOME = 1
    assert fpdf.SOME == 1

    import fpdf

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_DIR
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_DIR = "/tmp"
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_MODE
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    with pytest.warns(DeprecationWarning) as record:
        fpdf.FPDF_CACHE_MODE = 1
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    fpdf.SOME = 1
    assert fpdf.SOME == 1


def test_add_font_with_str_fname_ok(tmp_path):
    font_file_path = str(HERE / "Roboto-Regular.ttf")
    for font_cache_dir in (True, str(tmp_path), None):
        with pytest.warns(DeprecationWarning) as record:
            pdf = FPDF(font_cache_dir=font_cache_dir)
            pdf.add_font(fname=font_file_path)
            pdf.set_font("Roboto-Regular", size=64)
            pdf.add_page()
            pdf.cell(text="Hello World!")
            assert_pdf_equal(pdf, HERE / "add_font_unicode.pdf", tmp_path)

        for r in record:
            if r.category == DeprecationWarning:
                assert_same_file(r.filename, __file__)


def test_add_core_fonts():
    font_file_path = HERE / "Roboto-Regular.ttf"
    pdf = FPDF()
    pdf.add_page()

    with pytest.warns(UserWarning) as record:  # "already added".
        pdf.add_font("Helvetica", fname=font_file_path)
        pdf.add_font("Helvetica", style="B", fname=font_file_path)
        pdf.add_font("helvetica", style="IB", fname=font_file_path)
        pdf.add_font("times", style="", fname=font_file_path)
        pdf.add_font("courier", fname=font_file_path)
        assert not pdf.fonts  # No fonts added, as all of them are core fonts

    for r in record:
        assert_same_file(r.filename, __file__)


def test_render_en_dash(tmp_path):  # issue-166
    pdf = FPDF()
    pdf.add_font(fname=HERE / "Roboto-Regular.ttf")
    pdf.set_font("Roboto-Regular", size=120)
    pdf.add_page()
    pdf.cell(w=pdf.epw, text="–")  # U+2013
    assert_pdf_equal(pdf, HERE / "render_en_dash.pdf", tmp_path)


def test_add_font_otf(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Quicksand", style="", fname=HERE / "Quicksand-Regular.otf")
    pdf.add_font("Quicksand", style="B", fname=HERE / "Quicksand-Bold.otf")
    pdf.add_font("Quicksand", style="I", fname=HERE / "Quicksand-Italic.otf")
    pdf.set_font("Quicksand", size=32)
    text = (
        "Lorem ipsum dolor, **consectetur adipiscing** elit,"
        " eiusmod __tempor incididunt__ ut labore et dolore --magna aliqua--."
    )
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True)
    pdf.ln()
    pdf.multi_cell(w=pdf.epw, text=text, markdown=True, align="L")
    assert_pdf_equal(pdf, HERE / "fonts_otf.pdf", tmp_path)


def test_add_font_uppercase():
    pdf = FPDF()
    pdf.add_font(fname=HERE / "Roboto-BoldItalic.TTF")
    assert pdf.fonts is not None and len(pdf.fonts) != 0  # fonts add successful


def test_add_font_missing_notdef_glyph(caplog):
    pdf = FPDF()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular-without-notdef.ttf")
    assert pdf.fonts is not None and len(pdf.fonts) != 0  # fonts add successful
    assert (
        "TrueType Font 'roboto' is missing the '.notdef' glyph. "
        "Fallback glyph will be provided."
    ) in caplog.text


def test_font_missing_glyphs(caplog):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.set_font("Roboto")
    pdf.cell(text="Test 𝕥𝕖𝕤𝕥 🆃🅴🆂🆃 😲")
    pdf.output(devnull)
    assert (
        "Roboto is missing the following glyphs: "
        "'𝕥' (\\U0001d565), '𝕖' (\\U0001d556), '𝕤' (\\U0001d564), "
        "'🆃' (\\U0001f183), '🅴' (\\U0001f174), '🆂' (\\U0001f182), '😲' (\\U0001f632)"
        in caplog.text
    )


def test_font_with_more_than_10_missing_glyphs(caplog):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.set_font("Roboto")
    pdf.cell(
        text="Ogham space mark: '\U00001680' - Ideographic space: '\U00003000' - 𝒯ℰ𝒮𝒯 ⓉⒺⓈⓉ 𝕥𝕖𝕤𝕥 🆃🅴🆂🆃 🇹🇪🇸🇹"
    )
    pdf.output(devnull)
    assert (
        "Roboto is missing the following glyphs: "
        "' ' (\\u1680), '　' (\\u3000), "
        "'𝒯' (\\U0001d4af), 'ℰ' (\\u2130), '𝒮' (\\U0001d4ae), "
        "'Ⓣ' (\\u24c9), 'Ⓔ' (\\u24ba), 'Ⓢ' (\\u24c8), "
        "'𝕥' (\\U0001d565), '𝕖' (\\U0001d556), ... (and 7 others)" in caplog.text
    )


def test_add_font_woff(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Noto", style="", fname=HERE / "noto-sans-v42-latin-regular.woff")
    pdf.set_font("Noto", size=32)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "font_woff.pdf", tmp_path)


def test_add_font_woff_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Noto", style="", fname=HERE / "noto-sans-v42-latin-regular.woff")
    pdf.set_font("Noto", size=32)
    pdf.set_text_shaping(True)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "font_woff_hb.pdf", tmp_path)


def test_add_font_woff2(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Noto", style="", fname=HERE / "noto-sans-v42-latin-regular.woff2")
    pdf.set_font("Noto", size=32)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "font_woff2.pdf", tmp_path)


def test_add_font_woff2_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Noto", style="", fname=HERE / "noto-sans-v42-latin-regular.woff2")
    pdf.set_font("Noto", size=32)
    pdf.set_text_shaping(True)
    pdf.multi_cell(w=pdf.epw, text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "font_woff2_hb.pdf", tmp_path)


def test_add_font_woff2_without_brotli(monkeypatch):
    monkeypatch.setattr(woff2, "haveBrotli", False, raising=True)

    pdf = FPDF()
    with pytest.raises(
        RuntimeError,
        match=r"^Could not open WOFF2 font\. WOFF2 support requires an external Brotli",
    ):
        pdf.add_font("Noto", style="", fname=HERE / "noto-sans-v42-latin-regular.woff2")


def test_add_font_collection_all_faces(tmp_path):
    """
    This test will render all faces in the NotoSansCJK font collection.
    This font has multiple faces for the same glyphs, with regional variations
    for Chinese, Japanese, and Korean.
    The output PDF will show the glyphs with those slight regional variations.
    """
    COLLECTION_FONT = HERE / "noto-cjk-subset.ttc"
    COLLECTION_TEXT_ALL_FACES = "\u7e9b\u88ef\u8b56\u8c41\u904d\u98ef"
    pdf = FPDF()
    pdf.add_page()
    collection = TTCollection(str(COLLECTION_FONT))
    face_count = len(collection.fonts)
    for face_number in range(face_count):
        family = f"NotoCJKFace{face_number}"
        pdf.add_font(
            family=family,
            style="",
            fname=COLLECTION_FONT,
            collection_font_number=face_number,
        )
        pdf.set_font("helvetica", size=12)
        pdf.cell(text=f"Face {face_number}: ")
        pdf.set_font(family, size=12)
        pdf.cell(text=COLLECTION_TEXT_ALL_FACES)
        pdf.ln()
    assert_pdf_equal(pdf, HERE / "collection_all_faces.pdf", tmp_path)


def test_add_font_collection_shaping(tmp_path):
    COLLECTION_FONT = HERE / "noto-cjk-subset.ttc"
    COLLECTION_TEXT_ALL_FACES = "\u7e9b\u88ef\u8b56\u8c41\u904d\u98ef"
    pdf = FPDF()
    pdf.add_page()
    pdf.set_text_shaping(True)
    collection = TTCollection(str(COLLECTION_FONT))
    face_count = len(collection.fonts)
    for face_number in range(face_count):
        family = f"NotoCJKFace{face_number}"
        pdf.add_font(
            family=family,
            style="",
            fname=COLLECTION_FONT,
            collection_font_number=face_number,
        )
        pdf.set_font("helvetica", size=12)
        pdf.cell(text=f"Face {face_number}: ")
        pdf.set_font(family, size=12)
        pdf.cell(text=COLLECTION_TEXT_ALL_FACES)
        pdf.ln()
    assert_pdf_equal(pdf, HERE / "collection_all_faces_shaping.pdf", tmp_path)


SYMBOL_FONT_PATH = Path(r"c:\Windows\Fonts\symbol.ttf")


@pytest.mark.skipif(
    sys.platform not in ("cygwin", "win32"), reason="Windows-only symbol font test"
)
@pytest.mark.skipif(not SYMBOL_FONT_PATH.exists(), reason="symbol.ttf not available")
def test_add_font_symbol(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("MSSymbol", style="", fname=SYMBOL_FONT_PATH)
    pdf.set_font("MSSymbol", size=32)
    pdf.set_text_shaping(True)
    pdf.multi_cell(
        w=pdf.epw,
        text="ABCDEFGHIJKLMNOPQRSTUVWXYZ\nabcdefghijklmnopqrstuvwxyz\n0123456789 !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~",
    )
    assert_pdf_equal(pdf, HERE / "symbol.pdf", tmp_path)


@pytest.mark.skipif(
    sys.platform not in ("cygwin", "win32"), reason="Windows-only symbol font test"
)
@pytest.mark.skipif(not SYMBOL_FONT_PATH.exists(), reason="symbol.ttf not available")
def test_add_font_symbol_shaping(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("MSSymbol", style="", fname=SYMBOL_FONT_PATH)
    pdf.set_font("MSSymbol", size=32)
    pdf.set_text_shaping(True)
    pdf.multi_cell(
        w=pdf.epw,
        text="ABCDEFGHIJKLMNOPQRSTUVWXYZ\nabcdefghijklmnopqrstuvwxyz\n0123456789 !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~",
    )
    assert_pdf_equal(pdf, HERE / "symbol_shaping.pdf", tmp_path)



from pathlib import Path
from pytest import raises, warns
import qrcode
from fpdf.template import Template, FPDFException
from ..conftest import assert_pdf_equal
from test.template.charwrap_test_elements import elements as charwrap_elements

HERE = Path(__file__).resolve().parent


def test_template_nominal_hardcoded(tmp_path):
    """Taken from docs/Templates.md"""
    elements = [
        {
            "name": "company_logo",
            "type": "I",
            "x1": 20,
            "y1": 20,
            "x2": 70,
            "y2": 70,
            "font": None,
            "size": 0,
            "bold": 0,
            "italic": 0,
            "underline": 0,
            "foreground": 0,
            "text": "logo",
            "priority": 2,
        },
        {
            "name": "company_name",
            "type": "T",
            "x1": 20,
            "y1": 75,
            "x2": 118,
            "y2": 90,
            "font": "helvetica",
            "size": 12,
            "bold": 1,
            "italic": 1,
            "underline": 1,
            "foreground": 0,
            "text": "",
            "priority": 2,
            # multiline is optional, so we test some items without it.
        },
        {
            "name": "multiline_text",
            "type": "T",
            "x1": 20,
            "y1": 100,
            "x2": 40,
            "y2": 105,
            "font": "helvetica",
            "size": 12,
            "bold": 0,
            "italic": 0,
            "underline": 0,
            "foreground": 0,
            "background": 0x88FF00,
            "text": "Lorem ipsum dolor sit amet, consectetur adipisici elit",
            "priority": 2,
            "multiline": True,
        },
        {
            "name": "box",
            "type": "B",
            "x1": 15,
            "y1": 15,
            "x2": 185,
            "y2": 260,
            "font": "helvetica",
            "size": 0,
            "bold": 0,
            "italic": 0,
            "underline": 0,
            "foreground": 0,
            "text": None,
            "priority": 0,
        },
        {
            "name": "box_x",
            "type": "B",
            "x1": 95,
            "y1": 15,
            "x2": 105,
            "y2": 25,
            "font": "helvetica",
            "size": 0,
            "bold": 1,
            "italic": 0,
            "underline": 0,
            "foreground": 0,
            "text": None,
            "priority": 2,
        },
        {
            "name": "line1",
            "type": "L",
            "x1": 100,
            "y1": 25,
            "x2": 100,
            "y2": 57,
            "font": "helvetica",
            "size": 0,
            "bold": 0,
            "italic": 0,
            "underline": 0,
            "foreground": 0,
            "text": None,
            "priority": 3,
            "multiline": False,
        },
        {
            "name": "barcode",
            "type": "BC",
            "x1": 20,
            "y1": 246.5,
            "x2": 140,
            "y2": 254,
            "font": "Interleaved 2of5 NT",
            "size": 0.75,
            "bold": 0,
            "italic": 0,
            "underline": 0,
            "foreground": 0,
            "text": "200000000001000159053338016581200810081",
            "priority": 3,
            "multiline": False,
        },
    ]
    tmpl = Template(format="A4", elements=elements, title="Sample Invoice")
    tmpl.add_page()
    tmpl["company_name"] = "Sample Company"
    assert tmpl["company_name"] == "Sample Company"  # testing Template.__getitem__
    tmpl["company_logo"] = HERE.parent.parent / "docs/fpdf2-logo.png"
    assert_pdf_equal(tmpl, HERE / "template_nominal_hardcoded.pdf", tmp_path)


def test_template_nominal_json(tmp_path):
    """The contents of "nominal.json" must be exactly equivalent to the
    literal dict in test_template_nominal_hardcoded() above, since
    the same test file is used.
    When changes are necessary, ideally edit the JSON, then generate the Python
    list-of-dicts from that.
    """
    tmpl = Template(format="A4", title="Sample Invoice")
    tmpl.add_page()
    tmpl.parse_json(HERE / "nominal.json")
    tmpl["company_name"] = "Sample Company"
    assert tmpl["company_name"] == "Sample Company"  # testing Template.__getitem__
    tmpl["company_logo"] = HERE.parent.parent / "docs/fpdf2-logo.png"
    assert_pdf_equal(tmpl, HERE / "template_nominal_hardcoded.pdf", tmp_path)


def test_template_nominal_csv(tmp_path):
    """Same data as in docs/Templates.md
    The numeric_text tests for a regression."""
    tmpl = Template(format="A4", title="Sample Invoice")
    tmpl.parse_csv(HERE / "template_definition.csv", delimiter=";")
    tmpl.add_page()
    tmpl["empty_fields"] = "empty"
    assert_pdf_equal(tmpl, HERE / "template_nominal_csv.pdf", tmp_path)

    tmpl = Template(format="A4", title="Sample Invoice")
    tmpl.parse_csv(HERE / "template_definition.csv", delimiter=";", encoding="utf-8")
    tmpl.add_page()
    tmpl["empty_fields"] = "empty"
    assert_pdf_equal(tmpl, HERE / "template_nominal_csv.pdf", tmp_path)


def test_template_multipage(tmp_path):
    """Testing a Template() populating several pages."""
    tmpl = Template(format="A4", title="Sample Invoice")
    tmpl.parse_csv(HERE / "template_definition.csv", delimiter=";")
    tmpl.add_page()
    tmpl["name0"] = "Joe Doe"
    tmpl["title0"] = "Director"
    tmpl.add_page()
    tmpl["name0"] = "Jane Doe"
    tmpl["title0"] = "General Manager"
    tmpl.add_page()
    tmpl["name0"] = "Heinz Mustermann"
    tmpl["title0"] = "Worker"
    assert_pdf_equal(tmpl, HERE / "template_multipage.pdf", tmp_path)


def test_template_textstyles(tmp_path):
    """Testing bold, italic, underline in template and in tags."""
    elements = [
        {
            "name": "tb",
            "type": "T",
            "x1": 20,
            "y1": 20,
            "x2": 30,
            "y2": 25,
            "text": "text bold",
            "bold": True,
        },
        {
            "name": "ti",
            "type": "T",
            "x1": 20,
            "y1": 30,
            "x2": 30,
            "y2": 35,
            "text": "text italic",
            "italic": True,
        },
        {
            "name": "tu",
            "type": "T",
            "x1": 20,
            "y1": 40,
            "x2": 30,
            "y2": 45,
            "text": "text underline",
            "underline": True,
        },
        {
            "name": "tbiu",
            "type": "T",
            "x1": 20,
            "y1": 50,
            "x2": 30,
            "y2": 55,
            "text": "text all",
            "bold": True,
            "italic": True,
            "underline": True,
        },
        {
            "name": "wb",
            "type": "W",
            "x1": 20,
            "y1": 60,
            "x2": 30,
            "y2": 65,
            "text": "write bold",
            "bold": True,
        },
        {
            "name": "wi",
            "type": "W",
            "x1": 20,
            "y1": 70,
            "x2": 30,
            "y2": 75,
            "text": "write italic",
            "italic": True,
        },
        {
            "name": "wu",
            "type": "W",
            "x1": 20,
            "y1": 80,
            "x2": 30,
            "y2": 85,
            "text": "write underline",
            "underline": True,
        },
        {
            "name": "wbiu",
            "type": "W",
            "x1": 20,
            "y1": 90,
            "x2": 30,
            "y2": 95,
            "text": "write all",
            "bold": True,
            "italic": True,
            "underline": True,
        },
        {
            "name": "tbt",
            "type": "T",
            "x1": 20,
            "y1": 100,
            "x2": 30,
            "y2": 105,
            "text": "<B>text bold tags</B>",
        },
        {
            "name": "tit",
            "type": "T",
            "x1": 20,
            "y1": 110,
            "x2": 30,
            "y2": 115,
            "text": "<I>text italic tags</I>",
        },
        {
            "name": "tut",
            "type": "T",
            "x1": 20,
            "y1": 120,
            "x2": 30,
            "y2": 125,
            "text": "<U>text underline tags</U>",
        },
        {
            "name": "wbt",
            "type": "W",
            "x1": 20,
            "y1": 130,
            "x2": 30,
            "y2": 135,
            "text": "<B>write bold tags</B>",
        },
        {
            "name": "wit",
            "type": "W",
            "x1": 20,
            "y1": 140,
            "x2": 30,
            "y2": 145,
            "text": "<I>write italic tags</I>",
        },
        {
            "name": "wut",
            "type": "W",
            "x1": 20,
            "y1": 150,
            "x2": 30,
            "y2": 155,
            "text": "<U>write underline tags</U>",
        },
    ]
    tmpl = Template(elements=elements)
    tmpl.add_page()
    assert_pdf_equal(tmpl, HERE / "template_textstyles.pdf", tmp_path)


def test_template_item_access():
    """Testing Template() getitem/setitem."""
    elements = [
        {
            "name": "name",
            "type": "T",
            "x1": 20,
            "y1": 75,
            "x2": 30,
            "y2": 90,
            "text": "default text",
        }
    ]
    templ = Template(elements=elements)
    assert ("notthere" in templ) is False
    with raises(FPDFException):
        templ["notthere"] = "something"
    with raises(KeyError):
        # pylint: disable=pointless-statement
        templ["notthere"]
    defaultval = templ["name"]  # find in default data
    assert defaultval == "default text"
    templ["name"] = "new text"
    defaultval = templ["name"]  # find in text data
    assert defaultval == "new text"
    # bad type item access
    with raises(AssertionError):
        # pylint: disable=pointless-statement
        templ[7]
    with raises(AssertionError):
        templ[7] = 8
    with raises(AssertionError):
        # pylint: disable=pointless-statement
        7 in templ


def test_template_badinput():
    """Testing Template() with non-conforming definitions."""
    for arg in (
        "format",
        "orientation",
        "unit",
        "title",
        "author",
        "subject",
        "creator",
        "keywords",
    ):
        with raises(TypeError):
            Template(**{arg: 7})  # numeric instead of str
    elements = [{}]
    with raises(KeyError):
        tmpl = Template(elements=elements)
    elements = [{"name": "n", "type": "X"}]
    with raises(KeyError):
        tmpl = Template(elements=elements)
        tmpl.render()
    elements = [  # missing mandatory x2
        {
            "name": "n",
            "type": "T",
            "x1": 0,
            "y1": 0,
            "y2": 0,
            "text": "Hello!",
        }
    ]
    with raises(KeyError):
        tmpl = Template(elements=elements)
        tmpl["n"] = "hello"
        tmpl.render()
    elements = [  # malformed y2
        {
            "name": "n",
            "type": "T",
            "x1": 0,
            "y1": 0,
            "x2": 0,
            "y2": "x",
            "text": "Hello!",
        }
    ]
    with raises(TypeError):
        tmpl = Template(elements=elements)
        tmpl["n"] = "hello"
        tmpl.render()
    tmpl = Template()
    tmpl.add_page()
    with raises(FPDFException):
        tmpl.parse_csv(HERE / "mandmissing.csv", delimiter=";")
    with raises(ValueError):
        tmpl.parse_csv(HERE / "badint.csv", delimiter=";")
    with raises(ValueError):
        tmpl.parse_csv(HERE / "badfloat.csv", delimiter=";")
    with raises(KeyError):
        tmpl.parse_csv(HERE / "badtype.csv", delimiter=";")
        tmpl.render()
    with warns(DeprecationWarning):
        Template(infile="whatever")
    with raises(AttributeError):
        with warns(DeprecationWarning):
            tmpl = Template()
            tmpl.render(dest="whatever")
    with raises(KeyError):
        tmpl.parse_json(HERE / "mandmissing.json")
    with raises(TypeError):
        tmpl.parse_json(HERE / "badtype.json")
    with raises(ValueError):
        tmpl.parse_json(HERE / "badcolor.json")


def test_template_code39(tmp_path):  # issue-161
    elements = [
        {
            "name": "code39",
            "type": "C39",
            "x1": 40,
            "y1": 50,
            "y2": 70,
            "size": 1.5,
            "text": "*Code 39 barcode*",
            "priority": 1,
        },
    ]
    tmpl = Template(format="A4", title="Sample Code 39 barcode", elements=elements)
    tmpl.add_page()
    assert_pdf_equal(tmpl, HERE / "template_code39.pdf", tmp_path)


def test_template_code39_legacy(tmp_path):
    # check that old parameters still work
    # This uses the same values as above, and compares to the same file.
    elements = [
        {
            "name": "code39",
            "type": "C39",
            "x": 40,
            "y": 50,
            "w": 1.5,
            "h": 20,
            "text": "*Code 39 barcode*",
            "priority": 1,
        },
    ]
    with warns(DeprecationWarning):
        tmpl = Template(format="A4", title="Sample Code 39 barcode", elements=elements)
        tmpl.add_page()
        assert_pdf_equal(tmpl, HERE / "template_code39.pdf", tmp_path)


def test_template_code39_defaultheight(tmp_path):  # height <= 0 invokes default
    elements = [
        {
            "name": "code39",
            "type": "C39",
            "x1": 40,
            "y1": 50,
            "y2": 50,
            "size": 1.5,
            "text": "*Code 39 barcode*",
            "priority": 1,
        },
    ]
    tmpl = Template(format="A4", title="Sample Code 39 barcode", elements=elements)
    tmpl.add_page()
    assert_pdf_equal(tmpl, HERE / "template_code39_defaultheight.pdf", tmp_path)


def test_template_qrcode(tmp_path):  # issue-175
    elements = [
        {
            "name": "barcode_0",
            "type": "I",
            "x1": 50,
            "y1": 50,
            "x2": 100,
            "y2": 100,
            "priority": 0,
            "text": None,
        },
        {
            "name": "barcode_1",
            "type": "I",
            "x1": 150,
            "y1": 150,
            "x2": 200,
            "y2": 200,
            "priority": 0,
            "text": None,
        },
    ]
    tmpl = Template(format="letter", elements=elements)
    tmpl.add_page()
    tmpl["barcode_0"] = qrcode.make("Test 0").get_image()
    tmpl["barcode_1"] = qrcode.make("Test 1").get_image()
    assert_pdf_equal(tmpl, HERE / "template_qrcode.pdf", tmp_path)


def test_rect_background(tmp_path):  # issue-203
    elements = [
        {
            "name": "A rectangle",
            "type": "B",
            "background": 0x80FF00,
            "foreground": 0x80FF00,
            "text": None,
            "priority": 1,
            "x1": 50,
            "y1": 50,
            "x2": 150,
            "y2": 150,
        },
    ]
    tmpl = Template(format="A4", elements=elements)
    tmpl.add_page()
    assert_pdf_equal(tmpl, HERE / "template_rect_background.pdf", tmp_path)


def test_template_justify(tmp_path):  # issue-207
    elements = [
        {
            "name": "paragraph",
            "type": "T",
            "x1": 10,
            "y1": 15,
            "x2": 580,
            "y2": 45,
            "font": "helvetica",
            "size": 16,
            "align": "J",
            "text": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore"
            " et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi"
            " ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit"
            " esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident,"
            " sunt in culpa qui officia deserunt mollit anim id est laborum.",
            "priority": 1,
            "multiline": True,
        },
    ]
    tmpl = Template(format="A4", unit="pt", elements=elements)
    tmpl.add_page()
    assert_pdf_equal(tmpl, HERE / "template_justify.pdf", tmp_path)


def test_template_split_multicell():
    elements = [
        {
            "name": "multiline_text",
            "type": "T",
            "x1": 20,
            "y1": 100,
            "x2": 60,
            "y2": 105,
            "font": "helvetica",
            "size": 12,
            "text": "Lorem ipsum",
        }
    ]
    text = (
        "Lorem ipsum dolor sit amet, consetetur sadipscing elitr,"
        " sed diam nonumy eirmod tempor invidunt ut labore et dolore"
        " magna aliquyam erat, sed diam voluptua."
    )
    expected = [
        "Lorem ipsum dolor",
        "sit amet, consetetur",
        "sadipscing elitr, sed",
        "diam nonumy",
        "eirmod tempor",
        "invidunt ut labore et",
        "dolore magna",
        "aliquyam erat, sed",
        "diam voluptua.",
    ]
    tmpl = Template(format="A4", unit="pt", elements=elements)
    res = tmpl.split_multicell(text, "multiline_text")
    assert res == expected


def test_template_wrapmode(tmp_path):
    # Test that wrap mode can optionally be used to set wrapping using characters instead of words.
    tmpl = Template(elements=charwrap_elements)
    tmpl.add_page()
    assert_pdf_equal(tmpl, HERE / "template_wrapmode.pdf", tmp_path)



from pathlib import Path
from test.conftest import assert_pdf_equal

import pytest

from fpdf import FPDF
from fpdf.drawing import (
    Arc,
    BezierCurve,
    BoundingBox,
    Ellipse,
    HorizontalLine,
    Line,
    PaintedPath,
    Point,
    QuadraticBezierCurve,
    Rectangle,
    RelativeArc,
    RelativeBezierCurve,
    RelativeHorizontalLine,
    RelativeLine,
    RelativeQuadraticBezierCurve,
    RelativeVerticalLine,
    RoundedRectangle,
    Transform,
    VerticalLine,
)
from fpdf.enums import PathPaintRule

HERE = Path(__file__).resolve().parent


def rotate_around_center(bbox: BoundingBox) -> Transform:
    """
    Create a transformation that rotates the bounding box around its center.
    """
    center_x = (bbox.x0 + bbox.x1) / 2
    center_y = (bbox.y0 + bbox.y1) / 2
    return (
        Transform.translation(center_x, center_y)
        .rotate(45)
        .translate(-center_x + 150, -center_y)
    )


TRANSFORMS = {
    "identity": Transform.identity(),
    "scaled": Transform.scaling(2, 3),
    "rotated": rotate_around_center,
    "translated": Transform.translation(50, 100),
}


@pytest.mark.parametrize(
    "shape, start, expected_bbox",
    [
        (
            Line(pt=Point(50, 20)),
            Point(10, 20),
            BoundingBox(10, 20, 50, 20),
        ),
        (
            Line(pt=Point(30, 80)),
            Point(30, 10),
            BoundingBox(30, 10, 30, 80),
        ),
        (
            Line(pt=Point(100, 100)),
            Point(50, 50),
            BoundingBox(50, 50, 100, 100),
        ),
        (
            Line(pt=Point(50, 50)),
            Point(100, 100),
            BoundingBox(50, 50, 100, 100),
        ),
        (
            Line(pt=Point(5, 5)),
            Point(-10, -5),
            BoundingBox(-10, -5, 5, 5),
        ),
        (Line(pt=Point(42, 42)), Point(42, 42), BoundingBox(42, 42, 42, 42)),
        (
            RelativeLine(pt=Point(40, 0)),
            Point(10, 20),
            BoundingBox(10, 20, 50, 20),
        ),
        (
            RelativeLine(pt=Point(0, 60)),
            Point(30, 10),
            BoundingBox(30, 10, 30, 70),
        ),
        (
            RelativeLine(pt=Point(100, 100)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 100),
        ),
        (
            RelativeLine(pt=Point(-100, -100)),
            Point(100, 100),
            BoundingBox(0, 0, 100, 100),
        ),
        (
            RelativeLine(pt=Point(15, 10)),
            Point(-10, -5),
            BoundingBox(-10, -5, 5, 5),
        ),
        (
            RelativeLine(pt=Point(0, 0)),
            Point(42, 42),
            BoundingBox(42, 42, 42, 42),
        ),
        (HorizontalLine(x=50), Point(10, 20), BoundingBox(10, 20, 50, 20)),
        (HorizontalLine(x=20), Point(50, 15), BoundingBox(20, 15, 50, 15)),
        (
            HorizontalLine(x=42),
            Point(42, 42),
            BoundingBox(42, 42, 42, 42),
        ),
        (
            RelativeHorizontalLine(x=40),
            Point(10, 20),
            BoundingBox(10, 20, 50, 20),
        ),
        (
            RelativeHorizontalLine(x=-30),
            Point(20, 15),
            BoundingBox(-10, 15, 20, 15),
        ),
        (
            RelativeHorizontalLine(x=0),
            Point(42, 42),
            BoundingBox(42, 42, 42, 42),
        ),
        (VerticalLine(y=80), Point(30, 10), BoundingBox(30, 10, 30, 80)),
        (VerticalLine(y=20), Point(20, 50), BoundingBox(20, 20, 20, 50)),
        (VerticalLine(y=42), Point(42, 42), BoundingBox(42, 42, 42, 42)),
        (RelativeVerticalLine(y=50), Point(30, 10), BoundingBox(30, 10, 30, 60)),
        (
            RelativeVerticalLine(y=-30),
            Point(20, 50),
            BoundingBox(20, 20, 20, 50),
        ),
        (
            RelativeVerticalLine(y=0),
            Point(42, 42),
            BoundingBox(42, 42, 42, 42),
        ),
        (
            BezierCurve(c1=Point(30, 0), c2=Point(70, 0), end=Point(100, 0)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 0),
        ),
        (
            BezierCurve(c1=Point(0, 30), c2=Point(0, 70), end=Point(0, 100)),
            Point(0, 0),
            BoundingBox(0, 0, 0, 100),
        ),
        (
            BezierCurve(c1=Point(100, 0), c2=Point(0, 100), end=Point(100, 100)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 100),
        ),
        (
            BezierCurve(c1=Point(70, 100), c2=Point(30, 100), end=Point(0, 100)),
            Point(100, 100),
            BoundingBox(0, 100, 100, 100),
        ),
        (
            BezierCurve(c1=Point(150, 0), c2=Point(-150, 100), end=Point(100, 100)),
            Point(0, 0),
            BoundingBox(-3.808522398518157, 0.0, 100.0, 100.0),
        ),
        (
            BezierCurve(c1=Point(25, 25), c2=Point(50, 50), end=Point(75, 75)),
            Point(0, 0),
            BoundingBox(0, 0, 75, 75),
        ),
        (
            BezierCurve(c1=Point(42, 42), c2=Point(42, 42), end=Point(42, 42)),
            Point(42, 42),
            BoundingBox(42, 42, 42, 42),
        ),
        (
            RelativeBezierCurve(
                c1=Point(30, 60), c2=Point(60, 90), end=Point(100, 100)
            ),
            Point(0, 0),
            BoundingBox(0, 0, 100, 100),
        ),
        (
            RelativeBezierCurve(
                c1=Point(20, -20), c2=Point(-30, 30), end=Point(-50, -50)
            ),
            Point(25, 25),
            BoundingBox(-25.0, -25.0, 29.63395588138013, 25.0),
        ),
        (
            RelativeBezierCurve(c1=Point(0, 0), c2=Point(0, 0), end=Point(50, 50)),
            Point(0, 0),
            BoundingBox(0, 0, 50, 50),
        ),
        (
            RelativeBezierCurve(c1=Point(0, 0), c2=Point(0, 0), end=Point(0, 0)),
            Point(10, 10),
            BoundingBox(10, 10, 10, 10),
        ),
        (
            RelativeBezierCurve(c1=Point(50, 1), c2=Point(100, 2), end=Point(150, 1)),
            Point(0, 0),
            BoundingBox(0.0, 0.0, 150.0, 1.4142135623730951),
        ),
        (
            RelativeBezierCurve(c1=Point(1, 50), c2=Point(2, 100), end=Point(1, 150)),
            Point(0, 0),
            BoundingBox(0.0, 0.0, 1.4142135623730951, 150.0),
        ),
        (
            QuadraticBezierCurve(ctrl=Point(50, 100), end=Point(100, 0)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 50),
        ),
        (
            QuadraticBezierCurve(ctrl=Point(50, 100), end=Point(100, 0)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 50),
        ),
        (
            QuadraticBezierCurve(ctrl=Point(50, 0), end=Point(100, 0)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 0),
        ),
        (
            QuadraticBezierCurve(ctrl=Point(50, 50), end=Point(100, 100)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 100),
        ),
        (
            QuadraticBezierCurve(ctrl=Point(25, -100), end=Point(50, 0)),
            Point(0, 0),
            BoundingBox(0, -50, 50, 0),
        ),
        (
            QuadraticBezierCurve(ctrl=Point(42, 42), end=Point(42, 42)),
            Point(42, 42),
            BoundingBox(42, 42, 42, 42),
        ),
        (
            RelativeQuadraticBezierCurve(ctrl=Point(50, 100), end=Point(100, 0)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 50),
        ),
        (
            RelativeQuadraticBezierCurve(ctrl=Point(50, 0), end=Point(100, 0)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 0),
        ),
        (
            RelativeQuadraticBezierCurve(ctrl=Point(50, 50), end=Point(100, 100)),
            Point(0, 0),
            BoundingBox(0, 0, 100, 100),
        ),
        (
            RelativeQuadraticBezierCurve(ctrl=Point(25, -100), end=Point(50, 0)),
            Point(0, 0),
            BoundingBox(0, -50, 50, 0),
        ),
        (
            RelativeQuadraticBezierCurve(ctrl=Point(0, 0), end=Point(0, 0)),
            Point(42, 42),
            BoundingBox(42, 42, 42, 42),
        ),
        (
            RelativeQuadraticBezierCurve(ctrl=Point(-25, 25), end=Point(-50, -50)),
            Point(100, 100),
            BoundingBox(50, 50, 100, 106.25),
        ),
        (
            Arc(
                radii=Point(50, 50),
                rotation=0,
                large=False,
                sweep=True,
                end=Point(50, 50),
            ),
            Point(0, 0),
            BoundingBox(0, 0, 50, 50),
        ),
        (
            Arc(
                radii=Point(50, 50),
                rotation=0,
                large=True,
                sweep=False,
                end=Point(-50, 0),
            ),
            Point(50, 0),
            BoundingBox(-50, -50, 50, 0),
        ),
        (
            Arc(
                radii=Point(50, 30),
                rotation=45,
                large=False,
                sweep=True,
                end=Point(100, 0),
            ),
            Point(0, 0),
            BoundingBox(-5.405509724430832, -68.61529643011123, 99.99999989849465, 0),
        ),
        (
            Arc(
                radii=Point(0.1, 0.1),
                rotation=0,
                large=False,
                sweep=True,
                end=Point(0.1, 0),
            ),
            Point(0, 0),
            BoundingBox(0, -0.013397459621556158, 0.1, 0),
        ),
        (
            Arc(
                radii=Point(25, 40),
                rotation=0,
                large=False,
                sweep=False,
                end=Point(-40, 25),
            ),
            Point(0, 0),
            BoundingBox(
                -39.99999998902608, -8.416886842006127, 0.0, 24.999999894774596
            ),
        ),
        (
            RelativeArc(
                radii=Point(50, 50),
                rotation=0,
                large=False,
                sweep=True,
                end=Point(50, 50),
            ),
            Point(0, 0),
            BoundingBox(0, 0, 50, 50),
        ),
        (
            RelativeArc(
                radii=Point(50, 50),
                rotation=0,
                large=True,
                sweep=False,
                end=Point(-100, 0),
            ),
            Point(50, 0),
            BoundingBox(-50.0, -50.0, 50.0, 0.0),
        ),
        (
            RelativeArc(
                radii=Point(50, 30),
                rotation=45,
                large=False,
                sweep=True,
                end=Point(100, 0),
            ),
            Point(0, 0),
            BoundingBox(-5.405509724430832, -68.61529643011123, 99.99999989849465, 0),
        ),
        (
            RelativeArc(
                radii=Point(0.1, 0.1),
                rotation=0,
                large=False,
                sweep=True,
                end=Point(0.1, 0),
            ),
            Point(0, 0),
            BoundingBox(0, -0.013397459621556158, 0.1, 0),
        ),
        (
            RelativeArc(
                radii=Point(25, 40),
                rotation=0,
                large=False,
                sweep=False,
                end=Point(-40, 25),
            ),
            Point(0, 0),
            BoundingBox(-39.99999998902608, -8.416886842006127, 0, 24.999999894774596),
        ),
        (
            Rectangle(org=Point(10, 20), size=Point(80, 60)),
            None,
            BoundingBox(10, 20, 90, 80),
        ),
        (
            Rectangle(org=Point(0, 0), size=Point(100, 100)),
            None,
            BoundingBox(0, 0, 100, 100),
        ),
        (
            Rectangle(org=Point(-50, -25), size=Point(30, 45)),
            None,
            BoundingBox(-50, -25, -20, 20),
        ),
        (
            Rectangle(org=Point(42, 42), size=Point(0, 0)),
            None,
            BoundingBox(42, 42, 42, 42),
        ),
        (
            Rectangle(org=Point(5, 10), size=Point(100, 0.1)),
            None,
            BoundingBox(5, 10, 105, 10.1),
        ),
        (
            Rectangle(org=Point(5, 10), size=Point(0.1, 100)),
            None,
            BoundingBox(5, 10, 5.1, 110),
        ),
        (
            RoundedRectangle(
                org=Point(0, 0), size=Point(100, 50), corner_radii=Point(10, 10)
            ),
            None,
            BoundingBox(0, 0, 100, 50),
        ),
        (
            RoundedRectangle(
                org=Point(10, 20), size=Point(30, 40), corner_radii=Point(0, 0)
            ),
            None,
            BoundingBox(10, 20, 40, 60),
        ),
        (
            RoundedRectangle(
                org=Point(50, 50), size=Point(-30, -20), corner_radii=Point(5, 5)
            ),
            None,
            BoundingBox(20, 30, 50, 50),
        ),
        (
            RoundedRectangle(
                org=Point(0, 0), size=Point(20, 10), corner_radii=Point(30, 30)
            ),
            None,
            BoundingBox(0, 0, 20, 10),
        ),
        (
            RoundedRectangle(
                org=Point(0, 0), size=Point(0, 50), corner_radii=Point(10, 10)
            ),
            None,
            BoundingBox(0, 0, 0, 50),
        ),
        (
            RoundedRectangle(
                org=Point(0, 0), size=Point(80, 0), corner_radii=Point(10, 10)
            ),
            None,
            BoundingBox(0, 0, 80, 0),
        ),
        (
            RoundedRectangle(
                org=Point(30, 30), size=Point(-10, -10), corner_radii=Point(0, 0)
            ),
            None,
            BoundingBox(20, 20, 30, 30),
        ),
        (
            Ellipse(radii=Point(10, 10), center=Point(0, 0)),
            None,
            BoundingBox(-10, -10, 10, 10),
        ),
        (
            Ellipse(radii=Point(10, 10), center=Point(100, 200)),
            None,
            BoundingBox(90, 190, 110, 210),
        ),
        (
            Ellipse(radii=Point(25, 10), center=Point(50, 75)),
            None,
            BoundingBox(25, 65, 75, 85),
        ),
        (
            Ellipse(radii=Point(0, 20), center=Point(50, 50)),
            None,
            BoundingBox.empty(),
        ),
        (
            Ellipse(radii=Point(15, 0), center=Point(0, 0)),
            None,
            BoundingBox.empty(),
        ),
    ],
)
def test_shape_bounding_box(shape, start, expected_bbox):
    bbox, _ = shape.bounding_box(start)
    if not bbox.is_valid():
        assert not expected_bbox.is_valid()
    else:
        assert bbox == expected_bbox

    for _, tf in TRANSFORMS.items():
        transform = tf(expected_bbox) if callable(tf) else tf

        start = start if start else Point(0, 0)
        path = PaintedPath()
        path.move_to(start.x, start.y)
        path.add_path_element(shape)
        path.style.stroke_color = "#000000"
        path.style.paint_rule = PathPaintRule.FILL_NONZERO
        path.transform = transform

        bbox, _ = path.bounding_box(start=start)

        expected_scaled_bbox = expected_bbox.transformed(transform)
        if not bbox.is_valid():
            assert not expected_scaled_bbox.is_valid()
            return
        assert bbox == expected_scaled_bbox


def simple_path_absolute_relative():
    path = PaintedPath()
    path.move_to(10, 10)
    path.line_to(100, 10)
    path.line_relative(0, 40)
    path.line_to(10, 50)
    path.close()
    return path


def circle_and_rectangle():
    path = PaintedPath()
    path.circle(60, 60, 30)
    path.rectangle(20, 20, 40, 40)
    return path


def mixed_curves_and_lines():
    path = PaintedPath()
    path.move_to(0, 0)
    path.curve_to(30, 60, 60, 90, 100, 100)
    path.quadratic_curve_relative(20, -100, 50, 0)
    path.line_to(0, 0)
    path.close()
    return path


def arc_and_line():
    path = PaintedPath()
    path.move_to(50, 50)
    path.arc_relative(25, 25, 0, False, True, 50, 50)
    path.line_to(25, 75)
    return path


def rect_base():
    p = PaintedPath()
    p.style.fill_color = "#88aaee"
    p.rectangle(10, 10, 90, 40)  # bbox: (10,10)-(100,50)
    return p


def fill_only_rect_nonzero():
    p = rect_base()
    p.style.fill_color = "#88aaee"
    p.style.stroke_color = None
    p.style.paint_rule = PathPaintRule.FILL_NONZERO
    return p  # FILL_NONZERO, no stroke expansion


def stroke_only_rect_w4():
    p = rect_base()
    p.style.fill_color = None
    p.style.stroke_color = "#000000"
    p.style.stroke_width = 4  # expand by 2 in x/y
    return p  # STROKE


def stroke_fill_rect_w2():
    p = rect_base()
    p.style.fill_color = "#cccccc"
    p.style.stroke_color = "#000000"
    p.style.stroke_width = 2  # expand by 1 in x/y
    p.style.paint_rule = PathPaintRule.STROKE_FILL_NONZERO
    return p  # STROKE_FILL_NONZERO


def evenodd_fill_with_hole():
    p = PaintedPath()
    # outer rect (20,20)-(90,90)
    p.rectangle(20, 20, 70, 70)
    # inner rect (40,40)-(70,70) => hole
    p.rectangle(40, 40, 30, 30)
    p.style.fill_color = "#88aaee"
    p.style.stroke_color = None
    p.style.paint_rule = PathPaintRule.FILL_EVENODD
    return p  # FILL_EVENODD, bbox is outer rect only


def scaled_stroke_rect_w2():
    p = rect_base()
    p.transform = Transform.scaling(2.0, 0.5)
    p.style.fill_color = None
    p.style.stroke_color = "#000000"
    p.style.stroke_width = 2
    p.style.paint_rule = PathPaintRule.STROKE
    return p
    # Geom scales to (20,5)-(200,25); row_norms=(2,0.5); pad=(2,0.5)


def translated_fill_only():
    p = rect_base()
    p.transform = Transform.translation(15, -5)
    p.style.fill_color = "#88aaee"
    p.style.stroke_color = None
    p.style.paint_rule = PathPaintRule.FILL_NONZERO
    return p  # FILL_NONZERO, just shifted


def nested_scaled_circle_stroke2():
    p = PaintedPath()
    p.transform = Transform.scaling(1.5, 1.5)
    p.circle(20, 20, 10)  # base bbox (10,10)-(30,30) -> scaled to (15,15)-(45,45)
    p.style.fill_color = None
    p.style.stroke_color = "#000000"
    p.style.stroke_width = 2  # row_norms ~ (1.5,1.5); pad=(1.5,1.5)
    return p


def hairline_stroke_rect():
    p = rect_base()
    p.style.fill_color = None
    p.style.stroke_color = "#000000"
    p.style.stroke_width = 0  # hairline: no bbox expansion in user space
    return p


def zero_opacity_stroke_rect():
    p = rect_base()
    p.style.fill_color = None
    p.style.stroke_color = "#000000"
    p.style.stroke_width = 6
    p.style.stroke_opacity = 0.0  # invisible stroke => no expansion
    return p


def dont_paint_rect():
    p = rect_base()
    p.style.paint_rule = PathPaintRule.DONT_PAINT
    # bbox should remain purely geometric
    return p


@pytest.mark.parametrize(
    "test_id, path_builder, expected_bbox",
    [
        (
            "simple_path_absolute_relative",
            simple_path_absolute_relative,
            BoundingBox(9.5, 9.5, 100.5, 50.5),
        ),
        (
            "circle_and_rectangle",
            circle_and_rectangle,
            BoundingBox(19.5, 19.5, 90.5, 90.5),
        ),
        (
            "mixed_curves_and_lines",
            mixed_curves_and_lines,
            BoundingBox(-0.5, -0.5, 150.5, 100.5),
        ),
        (
            "arc_and_line",
            arc_and_line,
            BoundingBox(
                24.5, 39.14466094591647, 110.85533905932736, 100.50000004195078
            ),
        ),
        # paint-rule variations (no transforms)
        (
            "fill_only_nonzero",
            fill_only_rect_nonzero,
            BoundingBox(10.0, 10.0, 100.0, 50.0),
        ),
        ("stroke_only_w4", stroke_only_rect_w4, BoundingBox(8.0, 8.0, 102.0, 52.0)),
        ("stroke_fill_w2", stroke_fill_rect_w2, BoundingBox(9.0, 9.0, 101.0, 51.0)),
        (
            "evenodd_fill_with_hole",
            evenodd_fill_with_hole,
            BoundingBox(20.0, 20.0, 90.0, 90.0),
        ),
        # transforms
        (
            "scaled_stroke_rect_w2",
            scaled_stroke_rect_w2,
            BoundingBox(18.0, 4.5, 202.0, 25.5),
        ),
        (
            "translated_fill_only",
            translated_fill_only,
            BoundingBox(25.0, 5.0, 115.0, 45.0),
        ),
        (
            "nested_scaled_circle_stroke2",
            nested_scaled_circle_stroke2,
            BoundingBox(13.5, 13.5, 46.5, 46.5),
        ),
        # special stroke cases
        (
            "hairline_stroke_rect",
            hairline_stroke_rect,
            BoundingBox(10.0, 10.0, 100.0, 50.0),
        ),
        (
            "zero_opacity_stroke_rect",
            zero_opacity_stroke_rect,
            BoundingBox(10.0, 10.0, 100.0, 50.0),
        ),
        # explicit DONT_PAINT (no stroke expansion)
        ("dont_paint_rect", dont_paint_rect, BoundingBox(10.0, 10.0, 100.0, 50.0)),
    ],
)
def test_composite_painted_path(test_id, path_builder, expected_bbox, tmp_path):
    path = path_builder()
    bbox, _ = path.bounding_box(Point(0, 0))
    print(bbox, expected_bbox)
    assert bbox == expected_bbox

    pdf = FPDF()
    pdf.add_page()

    with pdf.drawing_context() as gc:
        gc.add_item(path)

    with pdf.drawing_context() as gc:
        # Draw bounding box in red
        bbox_path = PaintedPath()
        bbox_path.rectangle(bbox.x0, bbox.y0, bbox.x1 - bbox.x0, bbox.y1 - bbox.y0)
        bbox_path.style.stroke_color = "#ff0000"
        bbox_path.style.stroke_width = 0.1
        bbox_path.style.fill_color = None
        gc.add_item(bbox_path)

    assert_pdf_equal(
        pdf,
        HERE / f"generated_pdf/bounding_box_painted_path_{test_id}.pdf",
        tmp_path,
    )



# pylint: disable=no-member
import re
from pathlib import Path
import pypdf
from fpdf import FPDF

HERE = Path(__file__).resolve().parent


def get_used_fonts_in_page(page):
    content = page.get_contents()
    if isinstance(content, pypdf.generic.ArrayObject):
        content = b"".join([x.get_object().get_data() for x in content])
    else:
        content = content.get_data()
    font_refs = re.findall(rb"/F(\d+)", content)
    return {int(ref) for ref in font_refs}


def test_unused_fonts_not_included(tmp_path):
    pdf = FPDF()
    pdf.add_font("Roboto", fname=HERE / "Roboto-Regular.ttf")  # F1
    pdf.add_font("Roboto", fname=HERE / "Roboto-Bold.ttf", style="B")  # F2
    pdf.add_font("Roboto", fname=HERE / "Roboto-Italic.ttf", style="I")  # F3
    pdf.add_font("Roboto", fname=HERE / "Roboto-BoldItalic.TTF", style="BI")  # F4
    pdf.set_font("Roboto", size=12)

    pdf.add_page()
    pdf.multi_cell(w=pdf.epw, text="**Text in bold**", markdown=True)  # use F2

    pdf.add_page()
    pdf.multi_cell(w=pdf.epw, text="__Text in italic__", markdown=True)  # use F3

    pdf.add_page()
    pdf.multi_cell(
        w=pdf.epw,
        text="Regular text\n**Text in bold**\n__Text in italic__",
        markdown=True,
    )  # use F1, F2, F3

    output_path = tmp_path / "throwaway.pdf"
    pdf.output(output_path)

    reader = pypdf.PdfReader(output_path)
    assert len(reader.pages) == 3

    for page_num, page in enumerate(reader.pages, start=1):
        resources = page["/Resources"]
        fonts = resources.get("/Font", {})
        used_font_ids = get_used_fonts_in_page(page)
        for font_key in fonts:
            font_id = int(font_key[2:])  # /F1 -> 1
            assert (
                font_id in used_font_ids
            ), f"Page {page_num} contains unused font {font_key}"

        if page_num == 1:
            assert used_font_ids == {2}, "Page 1 should only use F2"
        elif page_num == 2:
            assert used_font_ids == {3}, "Page 2 should only use F3"
        elif page_num == 3:
            assert used_font_ids == {1, 2, 3}, "Page 3 should use F1, F2, F3"


def test_unused_added_font_not_included(tmp_path):
    pdf = FPDF()
    pdf.add_font("Roboto", fname=HERE / "Roboto-Regular.ttf")  # F1
    pdf.add_font("Roboto", fname=HERE / "Roboto-Bold.ttf", style="B")  # F2

    pdf.add_page()
    pdf.set_font("Roboto")
    pdf.cell(text="Hello")

    output_path = tmp_path / "throwaway.pdf"
    pdf.output(output_path)

    reader = pypdf.PdfReader(output_path)
    fonts = reader.pages[0]["/Resources"]["/Font"]
    assert "F2" not in fonts, "Unused font F2 should not be included"


def test_font_set_but_not_used(tmp_path):
    pdf = FPDF()
    pdf.add_font("Roboto", fname=HERE / "Roboto-Regular.ttf")  # F1
    pdf.add_page()
    pdf.set_font("Roboto")
    pdf.add_page()
    pdf.set_font("Helvetica")
    pdf.cell(text="Hello")

    output_path = tmp_path / "throwaway.pdf"
    pdf.output(output_path)

    reader = pypdf.PdfReader(output_path)
    page = reader.pages[0]
    resources = page.get("/Resources", {})
    page1_fonts = resources.get("/Font", {}) if isinstance(resources, dict) else {}
    assert not page1_fonts, "Page 1 should have no fonts as none were used"


def test_multiple_pages_font_usage(tmp_path):
    pdf = FPDF()
    pdf.add_font("Roboto", fname=HERE / "Roboto-Regular.ttf")  # F1
    pdf.add_font("Roboto", fname=HERE / "Roboto-Bold.ttf", style="B")  # F2

    # Page 1: Use F1
    pdf.add_page()
    pdf.set_font("Roboto")
    pdf.cell(text="Page 1")

    # Page 2: Use F2
    pdf.add_page()
    pdf.set_font(style="B")
    pdf.cell(text="Page 2")

    output_path = tmp_path / "throwaway.pdf"
    pdf.output(output_path)

    reader = pypdf.PdfReader(output_path)
    page1_fonts = reader.pages[0]["/Resources"]["/Font"]
    page2_fonts = reader.pages[1]["/Resources"]["/Font"]

    assert list(page1_fonts.keys()) == ["/F1"], "Page 1 should only have F1"
    assert list(page2_fonts.keys()) == ["/F2"], "Page 2 should only have F2"


def test_nested_context_font_usage_after_page_break(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", fname=HERE / "Roboto-Regular.ttf")  # F1
    pdf.add_font("Roboto", fname=HERE / "Roboto-BoldItalic.TTF", style="BI")  # F2
    pdf.add_font(fname=HERE / "DejaVuSans.ttf")  # F3
    pdf.add_font(fname=HERE / "Garuda.ttf")  # F4
    font_mapping = {
        1: "Roboto-Regular",
        2: "Roboto-BoldItalic",
        3: "DejaVuSans",
        4: "Garuda",
    }

    # Outer context A
    with pdf.local_context():
        pdf.set_font("Roboto", size=12)
        pdf.write(text="A1 Roboto-Regular\n")

        # Context B
        with pdf.local_context():
            pdf.set_font(style="BI", size=14)
            pdf.write(text="B1 Roboto-BoldItalic\n")

            # Context C
            with pdf.local_context():
                pdf.set_font("DejaVuSans", style="", size=16)
                pdf.write(text="C1 DejaVuSans\n")

                # Context D - will trigger page break
                with pdf.local_context():
                    pdf.set_font("Garuda", size=18)
                    # Generate enough text to force page break
                    long_text = "D1 " + "D2Garuda " * 250  # ~100 words
                    pdf.multi_cell(w=pdf.epw, text=long_text)  # page break

                # After break: C context resumes but writes nothing

            # After break: B context resumes
            pdf.write(text="B2 ")  # Should use Roboto-BoldItalic again

        # After break: A context resumes
        pdf.write(text="A2 ")  # Should use Roboto again

    pdf.output(tmp_path / "test_nested_context_font_usage_after_page_break.pdf")

    reader = pypdf.PdfReader(
        tmp_path / "test_nested_context_font_usage_after_page_break.pdf"
    )
    assert len(reader.pages) == 2, "The PDF produced should have 2 pages"

    page1 = reader.pages[0]
    page1_used_fonts = set(font_mapping[f] for f in get_used_fonts_in_page(page1))
    page1_used_fonts_str = "Fonts used: " + ", ".join(page1_used_fonts)
    assert len(page1_used_fonts) == 4, (
        "Page 1 should use all fonts - " + page1_used_fonts_str
    )

    page2 = reader.pages[1]
    page2_used_fonts = set(font_mapping[f] for f in get_used_fonts_in_page(page2))
    page2_used_fonts_str = "Fonts used: " + ", ".join(page2_used_fonts)
    assert page2_used_fonts == {"Roboto-Regular", "Roboto-BoldItalic", "Garuda"}, (
        "Page 2 should use 3 fonts - " + page2_used_fonts_str
    )

    page2_resources = page2["/Resources"].get("/Font", {})
    for font_key in page2_resources:
        font_id = int(font_key[2:])  # convert /F1 -> 1
        font_name = font_mapping[font_id]
        assert (
            font_name in page2_used_fonts
        ), f"page 2 resource includes unused font：{font_name}（F{font_id}）"



import logging
import pytest
import sys

from pathlib import Path

from fpdf import FPDF
from fpdf.enums import DocumentCompliance
from fpdf.errors import PDFAComplianceError
from test.conftest import EPOCH, assert_pdf_equal

HERE = Path(__file__).resolve().parent
DUMMY_TXT_FILE = b"fpdf2"


def minimal_pdf(compliance) -> bytearray:
    mpdf = FPDF(enforce_compliance=compliance)
    mpdf.creation_date = EPOCH
    mpdf.add_page()
    return mpdf.output()


@pytest.mark.parametrize("dc", [DocumentCompliance.PDFA_1B])
def test_pdfa1_forbids_embed(dc):
    pdf = FPDF(enforce_compliance=dc)
    pdf.add_page()
    with pytest.raises(PDFAComplianceError) as err:
        pdf.embed_file(bytes=DUMMY_TXT_FILE, basename="test.txt")
    assert (
        str(err.value)
        == f"Embedding files is not allowed for documents compliant with {dc.label}"
    )


@pytest.mark.skipif(
    sys.platform in ("cygwin", "win32") and sys.version_info[:2] == (3, 14),
    reason="Skipped on Windows with Python 3.14 due to zlib compressed data differences",
)
@pytest.mark.parametrize(
    "dc",
    [DocumentCompliance.PDFA_2B, DocumentCompliance.PDFA_2U, DocumentCompliance.PDFA_4],
)
def test_pdfa_2_and_4_allows_only_pdf_and_warns(tmp_path, caplog, dc):
    pdf = FPDF(enforce_compliance=dc)
    pdf.add_page()
    with caplog.at_level(logging.WARNING):
        pdf.embed_file(
            bytes=minimal_pdf(dc),
            basename="a.pdf",
            modification_date=EPOCH,
            mime_type="application/pdf",
        )
    # Warn that the embedded PDF must itself be PDF/A:
    assert any(
        f"{dc.label}: ensure the embedded PDF 'a.pdf' is itself PDF/A to remain compliant."
        in r.getMessage()
        for r in caplog.records
    )
    assert_pdf_equal(pdf, HERE / f"{dc.name.lower()}_embed_pdf.pdf", tmp_path)


@pytest.mark.parametrize(
    "dc",
    [DocumentCompliance.PDFA_2B, DocumentCompliance.PDFA_2U, DocumentCompliance.PDFA_4],
)
def test_pdfa_2_and_4_base_forbids_embed_other(dc):
    pdf = FPDF(enforce_compliance=dc)
    pdf.add_page()
    with pytest.raises(PDFAComplianceError) as err:
        pdf.embed_file(
            bytes=b"col1,col2\n1,2\n",
            basename="data.csv",
            modification_date=EPOCH,
            mime_type="text/csv",
            associated_file_relationship="Data",
        )
    assert (
        str(err.value)
        == f"{dc.label} permits embedding only PDF files, which must themselves be PDF/A."
    )


@pytest.mark.parametrize(
    "dc",
    [DocumentCompliance.PDFA_2B, DocumentCompliance.PDFA_2U, DocumentCompliance.PDFA_4],
)
def test_pdfa_2_non_pdf_rejected(dc):
    pdf = FPDF(enforce_compliance=dc)
    pdf.add_page()
    with pytest.raises(PDFAComplianceError) as err:
        pdf.embed_file(bytes=DUMMY_TXT_FILE, basename="test.txt")
    assert (
        str(err.value)
        == f"{dc.label} permits embedding only PDF files, which must themselves be PDF/A."
    )


@pytest.mark.parametrize("dc", [DocumentCompliance.PDFA_3B, DocumentCompliance.PDFA_3U])
def test_pdfa_3_embeds_txt_with_filespec_and_subtype(tmp_path, dc):
    pdf = FPDF(enforce_compliance=dc)
    pdf.add_page()
    pdf.embed_file(
        bytes=DUMMY_TXT_FILE,
        basename="test.txt",
        modification_date=EPOCH,
        mime_type="text/plain",
        associated_file_relationship="Data",
    )
    assert_pdf_equal(pdf, HERE / f"{str(dc).lower()}_embed_txt.pdf", tmp_path)


@pytest.mark.parametrize("dc", [DocumentCompliance.PDFA_4F, DocumentCompliance.PDFA_4E])
def test_pdfa_4_variants_embed_with_af(tmp_path, dc):
    pdf = FPDF(enforce_compliance=dc)
    pdf.add_page()
    pdf.embed_file(
        bytes=b"col1,col2\n1,2\n",
        basename="data.csv",
        modification_date=EPOCH,
        mime_type="text/csv",
        associated_file_relationship="Data",
    )
    assert_pdf_equal(pdf, HERE / f"{str(dc).lower()}_embed_csv.pdf", tmp_path)


def test_pdfa_4f_requires_embedded_file():
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_4F)
    pdf.add_page()
    with pytest.raises(PDFAComplianceError) as err:
        pdf.output()
    assert str(err.value) == "PDF/A-4F requires at least one embedded file"



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal
from PIL import Image

HERE = Path(__file__).resolve().parent


def test_png_indexed_no_transparency(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()

    # P images
    assert Image.open(HERE / "flower1.png").mode == "P", "img.mode is not P"

    pdf.set_image_filter("FlateDecode")
    pdf.image(HERE / "flower1.png", x=10, y=10, w=50, h=50)

    pdf.set_image_filter("JPXDecode")
    pdf.image(HERE / "flower1.png", x=80, y=10, w=50, h=50)

    pdf.set_image_filter("DCTDecode")
    pdf.image(HERE / "flower1.png", x=150, y=10, w=50, h=50)

    pdf.set_image_filter("LZWDecode")
    pdf.image(HERE / "flower1.png", x=10, y=150, w=50, h=50)

    # PA images
    img = Image.open(HERE / "flower1.png").convert("PA")
    assert img.mode == "PA", "img.mode is not PA"

    pdf.set_image_filter("FlateDecode")
    pdf.image(img, x=10, y=80, w=50, h=50)

    pdf.set_image_filter("DCTDecode")
    pdf.image(img, x=80, y=80, w=50, h=50)

    pdf.set_image_filter("JPXDecode")
    pdf.image(img, x=150, y=80, w=50, h=50)

    pdf.set_image_filter("LZWDecode")
    pdf.image(img, x=10, y=220, w=50, h=50)

    assert_pdf_equal(pdf, HERE / "image_png_indexed_no_transparency.pdf", tmp_path)


def test_png_indexed_transparency(tmp_path):
    def insert_alpha_channel_from_RGBA(img, path_png):
        # open the image as RGBA
        img_RGBA = Image.open(path_png).convert("RGBA")

        # extract the alpha channel
        alpha = img_RGBA.tobytes()[slice(3, None, 4)]

        # create an image that represent the alpha layer
        alpha_layer = Image.frombytes(mode="L", size=img_RGBA.size, data=alpha)

        # put the alpha layer into the original image
        img.putalpha(alpha_layer)

        return img

    pdf = fpdf.FPDF()
    pdf.add_page()

    # P images - info[transparency]: int
    assert isinstance(
        Image.open(HERE / "flower2.png").info["transparency"], int
    ), "info['transparency]' is not int"

    pdf.set_image_filter("FlateDecode")
    pdf.image(HERE / "flower2.png", x=10, y=10, w=50, h=90)

    pdf.set_image_filter("JPXDecode")
    pdf.image(HERE / "flower2.png", x=80, y=10, w=50, h=90)

    pdf.set_image_filter("DCTDecode")
    pdf.image(HERE / "flower2.png", x=150, y=10, w=50, h=90)

    # P images - info[transparency]: bytes
    assert isinstance(
        Image.open(HERE / "flower3.png").info["transparency"], bytes
    ), "info['transparency]' is not bytes"

    pdf.set_image_filter("FlateDecode")
    pdf.image(HERE / "flower3.png", x=10, y=110, w=50, h=90)

    pdf.set_image_filter("JPXDecode")
    pdf.image(HERE / "flower3.png", x=80, y=110, w=50, h=90)

    pdf.set_image_filter("DCTDecode")
    pdf.image(HERE / "flower3.png", x=150, y=110, w=50, h=90)

    # PA images
    img = Image.open(HERE / "flower2.png").convert("PA")
    img = insert_alpha_channel_from_RGBA(img, HERE / "flower2.png")

    pdf.set_image_filter("FlateDecode")
    pdf.image(img, x=10, y=210, w=50, h=90)

    pdf.set_image_filter("DCTDecode")
    pdf.image(img, x=80, y=210, w=50, h=90)

    pdf.set_image_filter("JPXDecode")
    pdf.image(img, x=150, y=210, w=50, h=90)

    pdf.add_page()

    pdf.set_image_filter("LZWDecode")
    pdf.image(HERE / "flower2.png", x=10, y=10, w=50, h=90)

    pdf.set_image_filter("LZWDecode")
    pdf.image(HERE / "flower3.png", x=10, y=110, w=50, h=90)

    pdf.set_image_filter("LZWDecode")
    pdf.image(img, x=10, y=210, w=50, h=90)

    assert_pdf_equal(pdf, HERE / "image_png_indexed_transparency.pdf", tmp_path)



from pathlib import Path

from test.conftest import ensure_exec_time_below, ensure_rss_memory_below

from fpdf import FPDF

HERE = Path(__file__).resolve().parent
PNG_FILE_PATHS = list((HERE / "image/png_images/").glob("*.png"))
PNG_FILE_PATHS.extend(
    file_path
    for file_path in (HERE / "image/png_test_suite/").glob("*.png")
    if not file_path.name.startswith("x")
)


@ensure_exec_time_below(seconds=12)
@ensure_rss_memory_below(mib=30)  # VERY dependent on the environment
def test_intense_image_rendering(tmp_path):
    pdf = FPDF()
    for _ in range(2000):
        pdf.add_page()
        for i, png_file_path in enumerate(PNG_FILE_PATHS):
            x = (i % 13) * 16
            y = (i // 13) * 16
            pdf.image(png_file_path, x=x, y=y)
    pdf.output(tmp_path / "out.pdf")



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_ebdt_cozette(tmp_path):
    """
    Cozette is a bitmat font with MIT license.
    See https://github.com/the-moonwitch/Cozette
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Cozette", "", str(HERE / "ebdt_cozette.otb"))
    pdf.set_font("Cozette", "", 40)
    pdf.cell(text="Hello World!", new_x="LMARGIN", new_y="NEXT")
    pdf.set_text_color(0, 0, 255)
    pdf.cell(text="Hello Blue World!", new_x="LMARGIN", new_y="NEXT")
    pdf.set_text_color(0, 255, 0)
    pdf.cell(text="Hello Green World!", new_x="LMARGIN", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "ebdt_cozette.pdf", tmp_path)


def test_ebdt_terminus(tmp_path):
    """
    Terminus is a bitmat font with SIL open font license.
    See https://files.ax86.net/terminus-ttf/
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Terminus", "", str(HERE / "ebdt_TerminusTTF-4.49.3.ttf"))
    pdf.set_font("Terminus", "", 40)
    pdf.cell(text="Hello World!", new_x="LMARGIN", new_y="NEXT")
    pdf.set_text_color(0, 0, 255)
    pdf.cell(text="Hello Blue World!", new_x="LMARGIN", new_y="NEXT")
    pdf.set_text_color(0, 255, 0)
    pdf.cell(text="Hello Green World!", new_x="LMARGIN", new_y="NEXT")

    assert_pdf_equal(pdf, HERE / "ebdt_terminus.pdf", tmp_path)



from pathlib import Path

import pytest

from fpdf import FPDF
from fpdf.drawing_primitives import DeviceRGB
from fpdf.enums import DocumentCompliance

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
FONT_DIR = HERE / ".." / "fonts"


def test_basic_pdfa(tmp_path):
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_3B)
    pdf.set_lang("en-US")
    pdf.set_title("Tutorial7")
    pdf.set_subject("Example for PDFA")
    pdf.set_author(["John Dow", "Jane Dow"])
    pdf.set_keywords(["Example", "Tutorial7", "PDF/A"])
    pdf.add_font(fname=FONT_DIR / "DejaVuSans.ttf")
    pdf.add_font("DejaVuSans", style="B", fname=FONT_DIR / "DejaVuSans-Bold.ttf")
    pdf.add_font("DejaVuSans", style="I", fname=FONT_DIR / "DejaVuSans-Oblique.ttf")
    pdf.add_page()
    pdf.set_font("DejaVuSans", style="B", size=20)
    pdf.write(text="Header")
    pdf.ln(20)
    pdf.set_font(size=12)
    pdf.write(text="Example text")
    pdf.ln(20)
    pdf.set_font(style="I")
    pdf.write(text="Example text in italics")
    assert_pdf_equal(pdf, HERE / "basic_pdfa.pdf", tmp_path)


def test_pdfa_font_fallback(tmp_path):
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_2B)
    pdf.add_page()
    pdf.add_font(family="Quicksand", fname=FONT_DIR / "Quicksand-Regular.otf")
    pdf.add_font(family="Roboto", fname=FONT_DIR / "Roboto-Regular.ttf")
    pdf.add_font(family="DejaVuSans", fname=FONT_DIR / "DejaVuSans.ttf")
    pdf.add_font(family="TwitterEmoji", fname=FONT_DIR / "TwitterEmoji.ttf")
    pdf.add_font(family="Waree", fname=FONT_DIR / "Waree.ttf")
    text = "Hello world / สวัสดีชาวโลก ทดสอบฟอนต์, 😄 😁 😆 😅 ✌"
    pdf.set_fallback_fonts(
        ["DejaVuSans", "Quicksand", "Waree", "TwitterEmoji", "Roboto"]
    )
    pdf.set_text_shaping(True)
    pdf.set_font("TwitterEmoji", size=20)
    pdf.cell(text=text, new_x="LMARGIN", new_y="NEXT")
    pdf.ln()
    pdf.set_font("Roboto", size=20)
    pdf.cell(text=text, new_x="LMARGIN", new_y="NEXT")
    assert_pdf_equal(pdf, HERE / "pdfa_fallback_fonts.pdf", tmp_path)


def test_pdfa_png_image(tmp_path):
    pdf = FPDF(enforce_compliance=DocumentCompliance.PDFA_3B)
    pdf.add_page()
    pdf.image(
        HERE / ".." / "image" / "png_images" / "ac6343a98f8edabfcc6e536dd75aacb0.png",
        x=0,
        y=0,
        w=0,
        h=0,
    )
    assert_pdf_equal(pdf, HERE / "pdfa_image_png.pdf", tmp_path)


@pytest.mark.parametrize(
    "dc",
    [
        DocumentCompliance.PDFA_1B,
        DocumentCompliance.PDFA_2B,
        DocumentCompliance.PDFA_2U,
        DocumentCompliance.PDFA_3B,
        DocumentCompliance.PDFA_3U,
        DocumentCompliance.PDFA_4,
        DocumentCompliance.PDFA_4E,
    ],
)
def test_pdfa_transparent_png(tmp_path, dc):
    pdf = FPDF(enforce_compliance=dc)
    pdf.set_page_background(DeviceRGB(r=0, g=0, b=0))
    pdf.set_compression(False)
    pdf.add_font(fname=FONT_DIR / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans", size=10)
    pdf.add_page()
    pdf.image(
        HERE / ".." / "image" / "png_images" / "e59ec0cfb8ab64558099543dc19f8378.png",
        x=0,
        y=0,
        w=0,
        h=0,
    )
    assert_pdf_equal(
        pdf,
        HERE / f"{dc.name.lower()}_transparent_png.pdf",
        tmp_path,
    )
